import pytest

from nasap_net.models import Assembly, AuxEdge, BindingSite, Bond, Component
from nasap_net.reaction_exploration_im.lib import UniqueComb, \
    extract_unique_site_combinations


@pytest.fixture
def MX2():
    """X0(0)-(0)M0(1)-(0)X1"""
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    return Assembly(
        components={'M0': M, 'X0': X, 'X1': X},
        bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]
    )


def test_single_site(MX2):
    binding_site_combs = [
        (BindingSite('M0', 0),),
        (BindingSite('M0', 1),),
        (BindingSite('X0', 0),),
        (BindingSite('X1', 0),),
    ]

    result = extract_unique_site_combinations(binding_site_combs, MX2)

    assert result == {
        UniqueComb(site_comb=(BindingSite('M0', 0),), duplication=2),
        UniqueComb(site_comb=(BindingSite('X0', 0),), duplication=2),
    }


def test_site_pairs(MX2):
    binding_site_combs = [
        (BindingSite('M0', 0), BindingSite('X0', 0)),
        (BindingSite('M0', 0), BindingSite('X1', 0)),
        (BindingSite('M0', 1), BindingSite('X0', 0)),
        (BindingSite('M0', 1), BindingSite('X1', 0)),
    ]

    result = extract_unique_site_combinations(binding_site_combs, MX2)

    assert result == {
        UniqueComb(
            site_comb=(BindingSite('M0', 0), BindingSite('X0', 0)),
            duplication=2),
        UniqueComb(
            site_comb=(BindingSite('M0', 0), BindingSite('X1', 0)),
            duplication=2),
    }


def test_site_triplets():
    #          X1
    #         (0)
    #          |
    #         (1)
    # X0(0)-(0)M0(2)-(0)L0(1)
    #         (3)
    #          |
    #         (0)
    #          L1
    #         (1)
    M = Component(
        kind='M', sites=[0, 1, 2, 3],
        aux_edges=[AuxEdge(0, 1), AuxEdge(1, 2), AuxEdge(2, 3), AuxEdge(3, 0)])
    L = Component(kind='L', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    cis_ML2X2 = Assembly(
        {'M0': M, 'L0': L, 'L1': L, 'X0': X, 'X1': X},
        [Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0),
         Bond('M0', 2, 'L0', 0), Bond('M0', 3, 'L1', 0)]
    )
    binding_site_combs = [
        (BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1)),  # trans
        (BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L1', 1)),  # cis
        (BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L0', 1)),  # trans
        (BindingSite('M0', 1), BindingSite('X1', 0), BindingSite('L1', 1)),  # cis
    ]

    result = extract_unique_site_combinations(binding_site_combs, cis_ML2X2)

    assert result == {
        UniqueComb(
            (BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L0', 1)),
            duplication=2),
        UniqueComb(
            (BindingSite('M0', 0), BindingSite('X0', 0), BindingSite('L1', 1)),
            duplication=2),
    }



from collections.abc import Mapping

from nasap_net import Assembly, BindsiteIdConverter, Component


def count_connected(
        assembly: Assembly, comp_id: str, target_comp_kind: str,
        comp_kind_to_obj: Mapping[str, Component]
        ) -> int:
    """Count the number of connected components of a certain kind."""
    id_converter = BindsiteIdConverter()
    num = 0

    for bindsite in assembly.get_bindsites_of_component(
            comp_id, comp_kind_to_obj):
        connected = assembly.get_connected_bindsite(bindsite)
        if connected is None:
            continue
        if target_comp_kind is not None:
            if assembly.get_component_kind_of_bindsite(connected) == target_comp_kind:
                num += 1
        else:
            num += 1

    return num


from collections.abc import Mapping

from cachetools import cached
from cachetools.keys import hashkey

from nasap_net import Assembly, Component

_cache = {}  # type: ignore

def clear_cache_for_enum_valid_ml_pairs():
    """Clear the cache used by `enum_valid_ml_pairs`."""
    _cache.clear()


def _cache_key(
        assembly_id: str | int,
        assembly: Assembly,
        metal_kind: str, leaving_kind: str,
        component_structures: Mapping[str, Component]
        ) -> tuple:
    return hashkey(assembly_id, metal_kind, leaving_kind)


@cached(cache=_cache, key=_cache_key)
def enum_valid_ml_pairs(
        assembly_id: str | int,  # only used for caching
        assembly: Assembly,
        metal_kind: str, leaving_kind: str,
        component_structures: Mapping[str, Component]
        ) -> list[tuple[str, str]]:
    """Enumerate valid metal-leaving pairs in an assembly.

    "Valid" means that the bindsites are connected to each other, and the
    component kinds of the metal and leaving bindsites are metal_kind and
    leaving_kind, respectively.

    Parameters
    ----------
    assembly_id : str
        The ID of the assembly. Used for caching.
    assembly : Assembly
        The assembly containing the bindsites.
    metal_kind : str
        The kind of the metal bindsite.
    leaving_kind : str
        The kind of the leaving bindsite.
    component_structures : Mapping[str, ComponentStructure]
        A mapping of component names to their structures.

    Returns
    -------
    list of tuple of str
        A list of tuples, each containing a metal bindsite and a leaving bindsite.

    Note
    -----
    The result is cached to avoid redundant computation.
    The key for caching is based on the assembly_id, metal_kind, and leaving_kind,
    i.e., assembly and component_structures are not used.
    Make sure to provide the same assembly_id for the same assembly, and not to
    change the component_structures between calls.
    """
    # Pairs of metal bindsites and leaving bindsites
    # that meet the following conditions:
    # - The bindsites are connected to each other
    # - The component kind of the metal bindsite is metal_kind
    # - The component kind of the leaving bindsite is leaving_kind
    valid_ml_pairs = []
    for metal_bs in assembly.get_all_bindsites_of_kind(
            metal_kind, component_structures):
        connected_bs = assembly.get_connected_bindsite(metal_bs)
        if connected_bs is None:  # metal bindsite is free
            continue
        if assembly.get_component_kind_of_bindsite(
                connected_bs) == leaving_kind:
            valid_ml_pairs.append((metal_bs, connected_bs))
    return valid_ml_pairs



from collections.abc import Iterable, Mapping
from typing import Literal, TypeAlias, overload

import networkx as nx

from nasap_net import Assembly, Component

LAYOUT_NAME_TO_FUNC = {
    'spring': nx.spring_layout,
}


PosMap2D: TypeAlias = Mapping[str, tuple[float, float]]
PosMap3D: TypeAlias = Mapping[str, tuple[float, float, float]]
PosDict2D: TypeAlias = dict[str, tuple[float, float]]
PosDict3D: TypeAlias = dict[str, tuple[float, float, float]]


@overload
def calc_positions(
        assembly: Assembly,
        component_structures: Mapping[str, Component],
        *,
        dimensions: Literal['2d'] = '2d',
        layout_name: Literal['spring'] = 'spring', 
        init_pos: PosMap2D = {},
        fixed: Iterable[str] | Literal['auto'] | None = 'auto',
        other_layout_kwargs: Mapping = {}
        ) -> PosDict2D:
    ...
@overload
def calc_positions(
        assembly: Assembly,
        component_structures: Mapping[str, Component],
        *,
        dimensions: Literal['3d'],
        layout_name: Literal['spring'] = 'spring', 
        init_pos: PosMap3D = {},
        fixed: Iterable[str] | Literal['auto'] | None = 'auto',
        other_layout_kwargs: Mapping = {}
        ) -> PosDict3D:
    ...
def calc_positions(
        assembly: Assembly,
        component_structures: Mapping[str, Component],
        *,
        dimensions: Literal['2d', '3d'] = '2d',
        layout_name: Literal['spring'] = 'spring', 
        init_pos: PosMap2D | PosMap3D | None = {},
        fixed: Iterable[str] | Literal['auto'] | None = 'auto',
        other_layout_kwargs: Mapping = {}
        ) -> PosDict2D | PosDict3D:
    """Calculates the positions of the nodes in the assembly graph."""
    layout_func = LAYOUT_NAME_TO_FUNC[layout_name]

    g = assembly.g_snapshot(component_structures)

    init_pos, fixed = format_init_pos_and_fixed(init_pos, fixed)
    if not init_pos:
        init_pos = None
    if not fixed:
        fixed = None

    if dimensions == '2d':
        return layout_func(
            g, pos=init_pos, fixed=fixed, dim=2, 
            **other_layout_kwargs)
    elif dimensions == '3d':
        return layout_func(
            g, pos=init_pos, fixed=fixed, dim=3, 
            **other_layout_kwargs)
    else:
        raise ValueError(
            f"Invalid value for 'dimensions': {dimensions}. "
            "It must be either '2d' or '3d'.")


@overload
def format_init_pos_and_fixed(
        init_pos: None,
        fixed: Iterable[str] | Literal['auto'] | None,
        ) -> tuple[None, None]: ...
@overload
def format_init_pos_and_fixed(
        init_pos: PosMap2D, fixed: None,
        ) -> tuple[PosDict2D, None]: ...
@overload
def format_init_pos_and_fixed(
        init_pos: PosMap3D, fixed: None,
        ) -> tuple[PosDict3D, None]: ...
@overload
def format_init_pos_and_fixed(
        init_pos: PosMap2D,
        fixed: Iterable[str] | Literal['auto'],
        ) -> tuple[PosDict2D, set[str]]: ...
@overload
def format_init_pos_and_fixed(
        init_pos: PosMap3D,
        fixed: Iterable[str] | Literal['auto'],
        ) -> tuple[PosDict3D, set[str]]: ...
def format_init_pos_and_fixed(init_pos, fixed):
    if init_pos is None:
        return None, None
    
    if fixed is None:
        return dict(init_pos), None
    
    if fixed == 'auto':
        return dict(init_pos), set(init_pos.keys())
    
    if isinstance(fixed, Iterable) and init_pos:
        # Since nodes without initial position cannot be fixed, 
        # exclude them from the fixed set. 
        # (Otherwise, nx.spring_layout will raise an error.)
        fixed = set(fixed) & init_pos.keys()
        return dict(init_pos), fixed
    
    raise ValueError(
        f"Invalid value for 'fixed': {fixed}. "
        "It must be either 'auto' or an iterable of node IDs.")



from dataclasses import dataclass
from typing import Generic

from nasap_net.types import A


@dataclass(frozen=True)
class Reaction(Generic[A]):
    init_assem_id: A
    entering_assem_id: A | None
    product_assem_id: A
    leaving_assem_id: A | None
    metal_bs: str
    leaving_bs: str
    entering_bs: str


@dataclass(frozen=True)
class _MLE:
    metal: str
    leaving: str
    entering: str



import pytest

from nasap_net import Assembly, Component
from nasap_net.reaction_classification.utils import count_connected


def test():
    MLX = Assembly({'M1': 'M', 'L1': 'L', 'X1': 'X'},
                   [('M1.a', 'L1.a'), ('M1.b', 'X1.a')])
    COMP_KIND_TO_OBJ = {
        'M': Component(['a', 'b']),
        'L': Component(['a', 'b']),
        'X': Component(['a'])
    }
    
    connected_num = count_connected(
        MLX, 'M1', 'L', COMP_KIND_TO_OBJ)
    
    assert connected_num == 1


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



"""Integration tests for reaction classification.

Each test utilizes the functions from nasap_net.reaction_classification.utils
to construct a rule for reaction classification. The rule is then used to 
create a ReactionClassifier object, which is used to classify reactions.
"""

import pytest

from nasap_net import (Assembly, Component, InterReactionRich,
                       IntraReactionRich, ReactionClassifier)
from nasap_net.reaction_classification.utils import inter_or_intra

ReactionEmbedded = IntraReactionRich | InterReactionRich


def test_classification_by_inter_or_intra():
    def rule(reaction: ReactionEmbedded):
        if inter_or_intra(reaction) == "intra":
            return "intra"
        return "inter"
    
    classifier = ReactionClassifier(rule)

    intra = IntraReactionRich(
        init_assem=Assembly(), product_assem=Assembly(),
        leaving_assem=Assembly(), 
        metal_bs='', leaving_bs='', entering_bs='',
        duplicate_count=0
        )
    result = classifier.classify(intra)
    assert result == "intra"

    inter = InterReactionRich(
        init_assem=Assembly(), entering_assem=Assembly(),
        product_assem=Assembly(), leaving_assem=Assembly(),
        metal_bs='', leaving_bs='', entering_bs='',
        duplicate_count=0
        )
    result = classifier.classify(inter)
    assert result == "inter"


@pytest.fixture
def MLX():
    return Assembly({'M1': 'M', 'L1': 'L', 'X1': 'X'},
                    [('M1.a', 'L1.a'), ('M1.b', 'X1.a')])

@pytest.fixture
def L():
    return Assembly({'L1': 'L'})

@pytest.fixture
def ML2():
    return Assembly({'M1': 'M', 'L1': 'L', 'L2': 'L'},
                    [('M1.a', 'L1.a'), ('M1.b', 'L2.a')])

@pytest.fixture
def X():
    return Assembly({'X1': 'X'})

@pytest.fixture
def X_to_L(MLX, L, ML2, X):
    """MLX + L -> ML2 + X (L-X exchange)"""
    return InterReactionRich(
        init_assem=MLX, entering_assem=L,
        product_assem=ML2, leaving_assem=X,
        metal_bs='M1.a', leaving_bs='X1.a', entering_bs='L1.a',
        duplicate_count=2  # 1 (dup. on MLX) * 2 (dup. on L)
    )

@pytest.fixture
def L_to_L(MLX, L):
    """MLX + L -> MLX + L (L-L exchange)"""
    return InterReactionRich(
        init_assem=MLX, entering_assem=L,
        product_assem=MLX, leaving_assem=L,
        metal_bs='M1.a', leaving_bs='L1.a', entering_bs='L1.a',
        duplicate_count=2  # 1 (dup. on MLX) * 2 (dup. on L)
    )

@pytest.fixture
def comp_kind_to_obj():
    return {
        'M': Component(['a', 'b']),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
    }


def test_classification_by_comp_kinds(X_to_L, L_to_L):
    def rule(reaction: ReactionEmbedded):
        entering = reaction.entering_kind
        leaving = reaction.leaving_kind
        return f'{leaving}_to_{entering}'
    
    classifier = ReactionClassifier(rule)

    result = classifier.classify(X_to_L)
    assert result == "X_to_L"

    result = classifier.classify(L_to_L)
    assert result == "L_to_L"


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections import defaultdict
from collections.abc import Hashable, Iterable
from typing import Generic, TypeVar

_T = TypeVar('_T', bound=Hashable)


class UnionFind(Generic[_T]):
    """Union-Find"""
    def __init__(self, nodes: Iterable[_T]):
        self.parent: dict[_T, _T] = {node: node for node in nodes}

    def find(self, x: _T) -> _T:
        """Find the root of the set containing x with path compression."""
        if self.parent[x] != x:
            self.parent[x] = self.find(self.parent[x])
        return self.parent[x]

    def union(self, x: _T, y: _T) -> None:
        """Union the sets containing x and y."""
        x_root = self.find(x)
        y_root = self.find(y)
        if x_root != y_root:
            self.parent[y_root] = x_root

    @property
    def root_to_elements(self) -> dict[_T, set[_T]]:
        """Map each root to the set of its elements."""
        root_to_elements = defaultdict(set)
        for element in self.parent:
            root = self.find(element)
            root_to_elements[root].add(element)
        return root_to_elements



import pytest

from nasap_net.bondset_enumeration import normalize_bondset_under_sym_ops


def test_1():
    BONDSET = {1}
    SYM_OPS = {'foo': {1: 2, 2: 1}}
    result = normalize_bondset_under_sym_ops(BONDSET, SYM_OPS)
    assert result == {1}


def test_2():
    BONDSET = {2}
    SYM_OPS = {'foo': {1: 2, 2: 1}}
    result = normalize_bondset_under_sym_ops(BONDSET, SYM_OPS)
    assert result == {1}


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest
import yaml

from nasap_net import AuxEdge, Component


def test_typical_usage():
    binding_sites = ['a', 'b', 'c', 'd']
    aux_edges = [('a', 'b', 'cis'), ('b', 'c', 'cis'),
                 ('c', 'd', 'cis'), ('d', 'a', 'cis')]
    M = Component(binding_sites, aux_edges)
    assert M.binding_sites == set(binding_sites)
    assert M.aux_edges == {AuxEdge(*edge) for edge in aux_edges}


def test_init():
    binding_sites = {'a', 'b'}
    M = Component(binding_sites)
    assert M.binding_sites == binding_sites
    assert M.aux_edges == set()


def test_init_with_binding_sites_of_type_tuple():
    binding_sites = ['a', 'b']
    M = Component(binding_sites)
    assert M.binding_sites == set(binding_sites)
    assert M.aux_edges == set()


def test_init_with_aux_edges():
    binding_sites = {'a', 'b'}
    aux_edges = {AuxEdge('a', 'b', 'cis')}
    M = Component(binding_sites, aux_edges)
    assert M.binding_sites == binding_sites
    assert M.aux_edges == aux_edges


def test_init_with_aux_edges_of_type_tuple():
    binding_sites = {'a', 'b'}
    aux_edges = [('a', 'b', 'cis')]
    M = Component(binding_sites, aux_edges)
    assert M.binding_sites == binding_sites
    assert M.aux_edges == {AuxEdge(*edge) for edge in aux_edges}


def test_init_with_empty_binding_sites():
    # Raises no error.
    M = Component(set())
    assert M.binding_sites == set()


def test_init_with_empty_aux_edges():
    # Raises no error.
    M = Component({'a', 'b'}, set())
    assert M.aux_edges == set()


def test_init_with_invalid_aux_edge():
    with pytest.raises(ValueError):
        # The binding site 'c' is not in the binding sites.
        Component({'a', 'b'}, {AuxEdge('a', 'c', 'cis')})


def test_eq():
    M1 = Component({'a', 'b'}, {AuxEdge('a', 'b', 'cis')})
    M2 = Component({'a', 'b'}, {AuxEdge('a', 'b', 'cis')})
    assert M1 == M2


def test_eq_with_different_binding_sites():
    M1 = Component({'a', 'b'}, {AuxEdge('a', 'b', 'cis')})
    M2 = Component({'a', 'c'}, {AuxEdge('a', 'c', 'cis')})
    assert M1 != M2


def test_eq_with_different_aux_edges():
    M1 = Component({'a', 'b'}, {AuxEdge('a', 'b', 'cis')})
    M2 = Component({'a', 'b'}, {AuxEdge('a', 'b', 'trans')})
    assert M1 != M2


def test_yaml_serialization():
    M = Component({'a', 'b'}, {AuxEdge('a', 'b', 'cis')})
    M_yaml = yaml.dump(M, default_flow_style=None)
    loaded_M = yaml.safe_load(M_yaml)
    assert M == loaded_M


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import networkx as nx
import pytest
import yaml

from nasap_net import Assembly, AuxEdge, Component


@pytest.fixture
def M_COMP() -> Component:
    return Component({'a', 'b', 'c', 'd'})


@pytest.fixture
def M_WITH_AUX_EDGES() -> Component:
    return Component(
        {'a', 'b', 'c', 'd'}, 
        {
            AuxEdge('a', 'b', 'cis'), AuxEdge('b', 'c', 'cis'),
            AuxEdge('c', 'd', 'cis'), AuxEdge('d', 'a', 'cis')
        })


@pytest.fixture
def L_COMP() -> Component:
    return Component({'a', 'b'})


@pytest.fixture
def X_COMP() -> Component:
    return Component({'a'})


def test_init_with_no_args() -> None:
    assembly = Assembly()
    assert assembly.comp_id_to_kind == {}
    assert assembly.bonds == set()


def test_init() -> None:
    components = {
        'M1': 'M', 'L1': 'L', 'X1': 'X', 'X2': 'X', 'X3': 'X'}
    bonds = {
        frozenset(['M1.a', 'X1.a']), frozenset(['M1.b', 'X1.a']),
        frozenset(['M1.c', 'X2.a']), frozenset(['M1.d', 'X3.a']),}
    assembly = Assembly(components, bonds)

    assert assembly.comp_id_to_kind == components
    assert assembly.bonds == set(bonds)


def test_init_with_name() -> None:
    assembly = Assembly(name='MX')
    assert assembly.name == 'MX'


def test_add_component() -> None:
    assembly = Assembly()
    assembly.add_component('M1', 'M')
    assembly.add_component('L1', 'L')
    assembly.add_components([('X1', 'X'), ('X2', 'X'), ('X3', 'X')])
    assembly.add_bond('M1.a', 'X1.a')
    assembly.add_bonds([
        ('M1.b', 'X1.a'), ('M1.c', 'X2.a'), ('M1.d', 'X3.a')])

    assert assembly.comp_id_to_kind == {
        'M1': 'M', 'L1': 'L', 'X1': 'X', 'X2': 'X', 'X3': 'X'}
    assert assembly.bonds == {
        frozenset(['M1.a', 'X1.a']), frozenset(['M1.b', 'X1.a']),
        frozenset(['M1.c', 'X2.a']), frozenset(['M1.d', 'X3.a'])}


def test_rough_g_snapshot():
    # Test for Issue #31: 
    # https://github.com/Hiraoka-Group/nasap-net/issues/31#issue-2617886112
    
    # In cases where the assembly has parallel bonds, i.e., multiple bonds
    # between the same pair of components, the rough_g_snapshot method 
    # should return a MultiGraph with multiple edges between the same pair
    # of nodes.

    ML_RING = Assembly({'M1': 'M', 'L1': 'L'},
                       [('M1.a', 'L1.a'), ('M1.b', 'L1.b')])

    rough_g = ML_RING.rough_g_snapshot
    
    assert isinstance(rough_g, nx.MultiGraph)
    assert [set(e) for e in rough_g.edges()] == [
        {'M1', 'L1'}, {'M1', 'L1'}]


def test_yaml():
    assembly = Assembly(
        {'M1': 'M', 'X1': 'X', 'X2': 'X'},
        [('M1.a', 'X1.a'), ('M1.b', 'X2.a')])

    assembly_yaml = yaml.dump(assembly)
    loaded_assembly = yaml.safe_load(assembly_yaml)

    assert assembly == loaded_assembly


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Mapping
from typing import Literal

import matplotlib.pyplot as plt
import networkx as nx

from nasap_net import Assembly, Component

from .color_map import make_complete_color_map

__all__ = ['draw_2d']


def draw_2d(
        assembly: Assembly,
        component_structures: Mapping[str, Component],
        positions: Mapping[str, tuple[float, float]],
        *,
        show: bool = True,
        # Sizes and colors
        core_node_size: int = 1600,
        bindsite_node_size: int = 400,
        component_colors: Mapping[str, str] = {},
        # Layout
        figsize: tuple[float, float] = (6, 6),
        x_lim: tuple[float, float] | None = None,
        y_lim: tuple[float, float] | None = None,
        # Labels
        node_labeling_mode: Literal[
            'component_kind', 'component_id', 'core_and_bindsite_ids',
            None] = 'component_kind',
        node_labeling_args: Mapping = {'font_size': 12},
        label_aux_edges: bool = True,
        aux_edge_labeling_args: Mapping = {'font_size': 12},
        # Others
        show_axes: bool = False,
        show_grid: bool = False,
        ) -> None:
    component_colors = make_complete_color_map(assembly, component_colors)

    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.set_position((0, 0, 1, 1))
    ax.set_aspect('equal')

    if x_lim is not None:
        ax.set_xlim(x_lim)
    if y_lim is not None:
        ax.set_ylim(y_lim)
    
    g = assembly.g_snapshot(component_structures)

    # ========= Draw the graph =========

    # Draw cores
    cores = list(assembly.iter_all_cores())

    core_fill_colors = [
        component_colors[assembly.get_component_kind_of_core(n)] 
        for n in cores]
    
    nx.draw_networkx_nodes(
        g, positions, ax=ax, nodelist=cores,
        node_color=core_fill_colors, node_size=core_node_size)

    # Draw bindsite
    bindsites = list(assembly.get_all_bindsites(component_structures))

    bindsite_edge_colors = [
        component_colors[assembly.get_component_kind_of_bindsite(n)] 
        for n in bindsites]
    
    nx.draw_networkx_nodes(
        g, positions, ax=ax, nodelist=bindsites,
        node_color='white', edgecolors=bindsite_edge_colors, linewidths=2,
        node_size=bindsite_node_size)
    
    # Draw core-to-bindsite edges
    for comp_id, comp_kind in assembly.comp_id_to_kind.items():
        color = component_colors[comp_kind]
        core = assembly.get_core_of_the_component(comp_id)
        bindsites_of_the_comp = assembly.get_bindsites_of_component(
            comp_id, component_structures)
        
        for bindsite in bindsites_of_the_comp:
            nx.draw_networkx_edges(
                g, positions, edgelist=[(core, bindsite)], edge_color=color, 
                style='solid', width=12)
    
    # Draw bonds
    for bindsite1, bindsite2 in assembly.bonds:
        nx.draw_networkx_edges(
            g, positions, edgelist=[(bindsite1, bindsite2)], edge_color='grey', 
            style=(0, (2, 1)), width=2)
        
    # Draw aux edges
    for aux_edge in assembly.iter_aux_edges(component_structures):
        nx.draw_networkx_edges(
            g, positions, edgelist=[(aux_edge.bindsite1, aux_edge.bindsite2)], 
            edge_color='black', style='dotted')
    
    # Draw the labels
    labels = {}
    match node_labeling_mode:
        case 'component_kind':
            for comp_id, comp_kind in assembly.comp_id_to_kind.items():
                core = assembly.get_core_of_the_component(comp_id)
                labels[core] = comp_kind
        case 'component_id':
            for comp_id in assembly.component_ids:
                core = assembly.get_core_of_the_component(comp_id)
                labels[core] = comp_id
        case 'core_and_bindsite_ids':
            for core in cores:
                labels[core] = core
            for bindsite in bindsites:
                labels[bindsite] = bindsite

    if node_labeling_mode is not None:
        nx.draw_networkx_labels(
            g, positions, labels=labels, **node_labeling_args)

    if label_aux_edges:
        edge_labels = {}
        for aux_edge in assembly.iter_aux_edges(component_structures):
            edge_labels[
                (aux_edge.bindsite1, aux_edge.bindsite2)] = aux_edge.aux_type
        nx.draw_networkx_edge_labels(
            g, positions, edge_labels=edge_labels, **aux_edge_labeling_args)

    # ========= Set the plot =========
    
    core_width_pt = core_node_size ** 0.5  # 40 (pt)
    inch_per_pt = 1 / 72  # 0.014 (inch/pt)
    px_per_inches = fig.dpi  # 100 (px/inch)
    core_width_px = core_width_pt * inch_per_pt * px_per_inches  # 40 * 1/72 * 100 = 56 (px)
    
    fig_size_px = fig.get_size_inches() * fig.dpi  # np.array([600, 600])
    fig_width_px = fig_size_px[0]  # 600 (px)
    fig_height_px = fig_size_px[1]  # 600 (px)

    data_x_min = min([x for x, _ in positions.values()])
    data_x_max = max([x for x, _ in positions.values()])
    data_y_min = min([y for _, y in positions.values()])
    data_y_max = max([y for _, y in positions.values()])

    data_x_range = data_x_max - data_x_min
    data_y_range = data_y_max - data_y_min

    if data_x_range > data_y_range:
        x_fig_width_px = fig_width_px
        y_fig_width_px = fig_width_px * data_y_range / data_x_range
    else:
        x_fig_width_px = fig_height_px * data_x_range / data_y_range
        y_fig_width_px = fig_height_px

    x_core_width_norm = core_width_px / x_fig_width_px
    if y_fig_width_px == 0:
        y_core_width_norm = 0
    else:
        y_core_width_norm = core_width_px / y_fig_width_px

    margin_scale = .8
    ax.margins(x_core_width_norm * margin_scale, y_core_width_norm * margin_scale)

    # show axis ticks
    ax.tick_params(left=True, bottom=True, labelleft=True, labelbottom=True)

    if not show_axes:
        ax.set_axis_off()
    if not show_grid:
        ax.grid(False)

    plt.tight_layout()

    # plt.axis('off')

    if show:
        plt.show()



from nasap_net import Assembly

__all__ = ['union_assemblies']


def union_assemblies(
        assembly1: Assembly, assembly2: Assembly,
        rename_prefix1: str | None = None,
        rename_prefix2: str | None = None
        ) -> Assembly:
    """Union two assemblies.

    Parameters
    ----------
    assembly1
        The first assembly.
    assembly2
        The second assembly.
    rename_map1
        A mapping from the component ids of the first assembly to the new
        component ids. If `None`, the component ids are not changed.
    rename_map2
        A mapping from the component ids of the second assembly to the new
        component ids. If `None`, the component ids are not changed.

    Returns
    -------
    Assembly
        The union of the two assemblies.
    
    Raises
    ------
    ValueError
        If the assemblies have overlapping nodes.

    Examples
    --------
    >>> from nasap_net import Assembly
    >>> from nasap_net.assem_class.assem import Assembly
    """
    if rename_prefix1 is not None:
        rename_map1 = {
            comp_id: f'{rename_prefix1}{comp_id}'
            for comp_id in assembly1.component_ids}
        assembly1 = assembly1.rename_component_ids(rename_map1)

    if rename_prefix2 is not None:
        rename_map2 = {
            comp_id: f'{rename_prefix2}{comp_id}'
            for comp_id in assembly2.component_ids}
        assembly2 = assembly2.rename_component_ids(rename_map2)

    if set(assembly1.component_ids) & set(assembly2.component_ids):
        raise ValueError('The assemblies have overlapping component IDs.')

    component_id_to_kind = (
        assembly1.comp_id_to_kind | assembly2.comp_id_to_kind)
    bonds = assembly1.bonds | assembly2.bonds

    return Assembly(component_id_to_kind, bonds)



from .binding_site_equivalence import UniqueComb, extract_unique_site_combinations, group_equivalent_node_combs
from .isomorphic_assembly_search import AssemblyNotFoundError, IsomorphicAssemblyFinder, light_signature
from .reaction_resolver import ReactionOutOfScopeError, ReactionResolver
from .separation import SeparatedIntoMoreThanTwoPartsError, separate_if_possible



from collections.abc import Iterable, Mapping
from typing import overload

from cachetools import cached
from cachetools.keys import hashkey

from nasap_net import Assembly, Component
from nasap_net.utils import group_equivalent_nodes_or_nodesets

from .cached_self_isomorphism_iteration import \
    iter_self_isomorphisms_with_cache

_cache = {}  # type: ignore

def clear_cache_for_compute_unique_bindsites_or_bindsite_sets():
    """Clear the cache used by `compute_unique_bindsites_or_bindsite_sets`."""
    _cache.clear()


def _cache_key(
        assembly_id: str | int,
        assembly: Assembly,
        bindsites_or_bindsite_sets: Iterable[str] | Iterable[tuple[str, ...]],
        component_structures: Mapping[str, Component],
        ) -> tuple:
    return hashkey(assembly_id, frozenset(bindsites_or_bindsite_sets))


@overload
def compute_unique_bindsites_or_bindsite_sets(
        assembly_id: str | int, assembly: Assembly,
        bindsites_or_bindsite_sets: Iterable[str],
        component_structures: Mapping[str, Component],
        ) -> list[tuple[str, int]]: ...
@overload
def compute_unique_bindsites_or_bindsite_sets(
        assembly_id: str | int, assembly: Assembly,
        bindsites_or_bindsite_sets: Iterable[tuple[str, str]],
        component_structures: Mapping[str, Component],
        ) -> list[tuple[tuple[str, str], int]]: ...
@overload
def compute_unique_bindsites_or_bindsite_sets(
        assembly_id: str | int, assembly: Assembly,
        bindsites_or_bindsite_sets: Iterable[tuple[str, str, str]],
        component_structures: Mapping[str, Component],
        ) -> list[tuple[tuple[str, str, str], int]]: ...
@overload
def compute_unique_bindsites_or_bindsite_sets(
        assembly_id: str | int, assembly: Assembly,
        bindsites_or_bindsite_sets: Iterable[tuple[str, ...]],
        component_structures: Mapping[str, Component],
        ) -> list[tuple[tuple[str, ...], int]]: ...
@cached(cache=_cache, key=_cache_key)
def compute_unique_bindsites_or_bindsite_sets(
        assembly_id,  # only used for caching
        assembly, bindsites_or_bindsite_sets, 
        component_structures):
    """Compute unique bindsites or bindsite sets.

    Parameters
    ----------
    assembly_id : str
        The ID of the assembly. Used for caching.
    assembly : Assembly
        The assembly containing the bindsites.
    bindsites_or_bindsite_sets : Iterable[str] | Iterable[tuple[str, ...]]
        The bindsites or bindsite sets to be considered.
    component_structures : Mapping[str, ComponentStructure]
        A mapping of component names to their structures.

    Returns
    -------
    list of tuple[str, int] | tuple[tuple[str, ...], int]
        A list of unique bindsites or bindsite sets and their counts.

    Note
    -----
    The result is cached to avoid redundant computation.
    The key for caching is based on the assembly_id and the set of bindsites or
    bindsite sets, i.e., assembly and component_structures are not used.
    Make sure to provide the same assembly_id for the same assembly, and not to
    change the component_structures between calls.
    """
    isomorphisms = iter_self_isomorphisms_with_cache(
        assembly_id, assembly, component_structures)

    grouped_nodesets = group_equivalent_nodes_or_nodesets(
        bindsites_or_bindsite_sets, isomorphisms)
    
    return [
        (min(group), len(group)) for group  # type: ignore
        in grouped_nodesets]



import pytest

from nasap_net.pipelines.bondset_enumeration import validate_bonds


def test_no_duplicates():
    bonds = [1, 2, 3, 4]
    validate_bonds(bonds)


def test_with_duplicates():
    bonds = [1, 2, 2, 4]
    with pytest.raises(
            ValueError, match='"bonds" must not contain duplicates.'):
        validate_bonds(bonds)


if __name__ == '__main__':
    pytest.main(['-v', __file__])


import pytest

from nasap_net import Assembly
from nasap_net.reaction_classification.utils import find_shortest_path


def test():
    assembly = Assembly()
    assembly.add_components([('M0', 'M'), ('L0', 'L'), ('X0', 'X')])
    assembly.add_bonds([('M0.a', 'L0.a'), ('M0.b', 'X0.a')])

    assert find_shortest_path(assembly, 'M0', 'X0') == ['M0', 'X0']
    assert find_shortest_path(assembly, 'M0', 'L0') == ['M0', 'L0']
    assert find_shortest_path(assembly, 'L0', 'X0') == ['L0', 'M0', 'X0']


def test_no_path():
    assembly = Assembly()
    assembly.add_components([('M0', 'M'), ('L0', 'L'), ('X0', 'X')])
    assembly.add_bonds([('M0.a', 'L0.a')])

    assert find_shortest_path(assembly, 'M0', 'X0') is None
    assert find_shortest_path(assembly, 'M0', 'L0') == ['M0', 'L0']
    assert find_shortest_path(assembly, 'L0', 'X0') is None


def test_not_in_assembly():
    assembly = Assembly()
    assembly.add_components([('M0', 'M'), ('L0', 'L'), ('X0', 'X')])
    assembly.add_bonds([('M0.a', 'L0.a')])

    with pytest.raises(ValueError):
        find_shortest_path(assembly, 'M0', 'Y0')
    with pytest.raises(ValueError):
        find_shortest_path(assembly, 'Y0', 'M0')


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from collections.abc import Hashable, Iterable, Mapping
from typing import TypeVar

__all__ = ['apply_symmetry_operation']

_T = TypeVar('_T', bound=Hashable)


def apply_symmetry_operation(
        bondset: Iterable[_T],
        sym_op: Mapping[_T, _T]
        ) -> set[_T]:
    """Apply a symmetry operation to an assembly."""
    return set(sym_op[bond] for bond in bondset)



from nasap_net.isomorphism import is_isomorphic
from nasap_net.models import Assembly, Bond, Component


def test():
    M = Component(kind='M', sites=[0, 1])
    L = Component(kind='L', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MLX = Assembly(
        components={'M1': M, 'L1': L, 'X1': X},
        bonds=[Bond('M1', 0, 'L1', 0), Bond('M1', 1, 'X1', 0)]
    )
    MLX_permuted = Assembly(
        components={'L1': L, 'X1': X, 'M1': M},
        bonds=[Bond('M1', 1, 'X1', 0), Bond('M1', 0, 'L1', 0)]
    )
    assert is_isomorphic(MLX, MLX_permuted)


def test_same_component_but_different_objects():
    M = Component(kind='M', sites=[1, 2])
    X = Component(kind='X', sites=[1])
    MX = Assembly(
        components={'M1': M, 'X1': X}, bonds=[Bond('M1', 1, 'X1', 1)])

    ANOTHER_M = Component(kind='M', sites=[10, 20])
    ANOTHER_X = Component(kind='X', sites=[10])
    ANOTHER_MX = Assembly(
        components={'M2': ANOTHER_M, 'X2': ANOTHER_X},
        bonds=[Bond('M2', 10, 'X2', 10)])

    assert is_isomorphic(MX, ANOTHER_MX)


def test_different_component_kinds():
    """Components with different kinds should be considered as different,
    even if their structures are identical.
    """
    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MX = Assembly(
        components={'M1': M, 'X1': X}, bonds=[Bond('M1', 0, 'X1', 0)])

    FAKE_M = Component(kind='ANOTHER_KIND', sites=[0, 1])
    FAKE_MX = Assembly(
        components={'M1': FAKE_M, 'X1': X}, bonds=[Bond('M1', 0, 'X1', 0)])

    assert not is_isomorphic(MX, FAKE_MX)



import pytest

from nasap_net.algorithms import sort_bondsets, sort_bondsets_and_bonds

# TODO: Add more test cases.


def test_sort_bondsets():
    BOND_SUBSETS = [
        ['1', '3'],
        ['2', '1'],
        ['2', '3', '1'],
        ['1', '2', '3'],
        ['1'],
        ['2'],
    ]
    EXPECTED = [
        ['1'],
        ['2'],
        ['2', '1'],
        ['1', '3'],
        ['2', '3', '1'],
        ['1', '2', '3'],
    ]
    # Notes:
    # - ['2', '1'] < ['1', '3'] because they are sorted before comparison.
    # - ['2', '3', '1'] < ['1', '2', '3'] because they are equal after sorting,
    # and the order of the equal elements is preserved.
    assert sort_bondsets(BOND_SUBSETS) == EXPECTED


def test_sort_bondsets_and_bonds():
    BOND_SUBSETS = [
        ['1', '3'],
        ['2', '1'],
        ['2', '3', '1'],
        ['1', '2', '3'],
        ['1'],
        ['2'],
    ]
    EXPECTED = [
        ['1'],
        ['2'],
        ['1', '2'],
        ['1', '3'],
        ['1', '2', '3'],
        ['1', '2', '3'],
    ]
    # Bonds in each bondset are also sorted.
    assert sort_bondsets_and_bonds(BOND_SUBSETS) == EXPECTED


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import os
from typing import Any

import yaml


def read_file(
        input_path: os.PathLike | str,
        *,
        verbose: bool = False
        ) -> Any:
    """Read the input file."""
    if verbose:
        print(f'Reading from "{input_path}"...')
    with open(input_path, 'r') as f:
        data = yaml.safe_load(f)
    if verbose:
        print(f'Finished reading from "{input_path}".')
    return data



import pytest

from nasap_net.duplicate_exclusion import group_assemblies_by_isomorphism


def test_group_duplicates_by_isomorphism():
    pass


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Mapping

from cachetools import cached
from cachetools.keys import hashkey

from nasap_net import Assembly, Component

_cache = {}  # type: ignore

def clear_cache_for_enum_valid_entering_bindsites():
    """Clear the cache used by `enum_valid_entering_bindsites`."""
    _cache.clear()


def _cache_key(
        assembly_id: str | int,
        assembly: Assembly,
        entering_kind: str,
        component_structures: Mapping[str, Component],
        ) -> tuple:
    return hashkey(assembly_id, entering_kind)


@cached(cache=_cache, key=_cache_key)
def enum_valid_entering_bindsites(
        assembly_id: str | int,  # only used for caching
        assembly: Assembly,
        entering_kind: str,
        component_structures: Mapping[str, Component],
        ) -> list[str]:
    """Enumerate valid entering bindsites.

    "Valid" here means that the bindsite is free and the component kind
    of the bindsite is entering_kind.
    
    Parameters
    ----------
    assembly_id : str
        The ID of the assembly. Used for caching.
    assembly : Assembly
        The assembly containing the bindsites.
    entering_kind : str
        The kind of the entering bindsite.
    component_structures : Mapping[str, ComponentStructure]
        A mapping of component names to their structures.
    
    Returns
    -------
    list of str
        A list of valid entering bindsites.

    Note
    -----
    The result is cached to avoid redundant computation.
    The key for caching is based on the assembly_id, and entering_kind, 
    i.e., assembly and component_structures are not used.
    Make sure to provide the same assembly_id for the same assembly, and not to
    change the component_structures between calls.
    """
    # Entering bindsites that meet the following conditions:
    # - The bindsite is free
    # - The component kind of the entering bindsite is entering_kind
    entering_bindsites = [
        bs for bs in assembly.get_all_bindsites_of_kind(
        entering_kind, component_structures)
        if assembly.is_free_bindsite(bs)]
    return entering_bindsites



from collections.abc import Iterable, Mapping
from itertools import chain, product
from typing import Iterator, TypeVar

from nasap_net.models import Assembly
from nasap_net.types import ID
from .explorer import InterReactionExplorer, IntraReactionExplorer
from .lib import ReactionOutOfScopeError, ReactionResolver
from .models import MLEKind, Reaction

_T = TypeVar('_T', bound=ID)

def explore_reactions(
        assemblies: Mapping[_T, Assembly],
        mle_kinds: Iterable[MLEKind],
        ) -> Iterator[Reaction]:
    # Add assembly IDs to assemblies
    assems_with_ids = [
        assem.copy_with(id_=assem_id)
        for assem_id, assem in assemblies.items()]

    reaction_iters: list[Iterator[Reaction]] = []
    for mle_kind in mle_kinds:
        # Intra-molecular reactions
        for assem in assems_with_ids:
            intra_explorer = IntraReactionExplorer(assem, mle_kind)
            reaction_iters.append(intra_explorer.explore())

        # Inter-molecular reactions
        for init_assem, entering_assem in product(assems_with_ids, repeat=2):
            inter_explorer = InterReactionExplorer(
                init_assem, entering_assem, mle_kind)
            reaction_iters.append(inter_explorer.explore())

    resolver = ReactionResolver(assems_with_ids)

    for reaction in chain.from_iterable(reaction_iters):
        try:
            yield resolver.resolve(reaction)
        except ReactionOutOfScopeError:
            continue



import yaml

from nasap_net.models import AuxEdge, Component


def load_components(yaml_str: str) -> dict[str, Component]:
    """Load components from a YAML string."""
    return yaml.load(yaml_str, Loader=_ComponentLoader)  # type: ignore


class _ComponentLoader(yaml.SafeLoader):
    def ignore_aliases(self, _):
        return True


def _component_constructor(
        loader: _ComponentLoader,
        node: yaml.Node,
) -> Component:
    assert isinstance(node, yaml.MappingNode)
    mapping = loader.construct_mapping(node, deep=True)
    kind = mapping['kind']
    sites = mapping['sites']
    aux_edges = []
    for m in mapping.get('aux_edges', []):
        site1, site2 = m['sites']
        aux_kind = m.get('kind')
        aux_edges.append(AuxEdge(site1, site2, kind=aux_kind))
    return Component(kind=kind, sites=sites, aux_edges=aux_edges)


yaml.add_constructor(
    '!Component', _component_constructor, Loader=_ComponentLoader
)



import os

import pytest
import yaml
from click.testing import CliRunner

from nasap_net import Assembly, Component, is_isomorphic
from nasap_net.cli.main import main


def test_version():
    runner = CliRunner()
    result = runner.invoke(main, ['--version'])
    assert result.exit_code == 0
    assert 'nasap_net' in result.output


def test_enumerate_assemblies(tmp_path):
    runner = CliRunner()

    INPUT_DATA = {
        'bonds': [1, 2, 3, 4],
        'bond_adjacency': {
            1: [2],
            2: [1, 3],
            3: [2, 4],
            4: [3],
        },
        'sym_ops': {
            'C2': {1: 4, 2: 3, 3: 2, 4: 1}
        },
        'component_kinds': {
            'L': Component(['a', 'b']),
            'M': Component(['a', 'b']),
            'X': Component(['a']),
        },
        'components_and_their_kinds': {
            'M1': 'M',
            'M2': 'M',
            'L1': 'L',
            'L2': 'L',
            'L3': 'L'
        },
        'bonds_and_their_binding_sites': {
            1: ['L1.b', 'M1.a'],
            2: ['M1.b', 'L2.a'],
            3: ['L2.b', 'M2.a'],
            4: ['M2.b', 'L3.a']
        },
        'capping_config': {
            'target_component_kind': 'M',
            'capping_component_kind': 'X',
            'capping_binding_site': 'a'
        }
    }

    EXPECTED = {
        # L1--M1--X1
        0: Assembly(
            {'L1': 'L', 'M1': 'M', 'X1': 'X'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'X1.a')]),
        # L1--M1--L2
        1: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a')]),
        # X1--M1--L2--M2--X2
        2: Assembly(
            {'X1': 'X', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'X2': 'X'}, 
            [('X1.a', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'),
             ('M2.b', 'X2.a')]),
        # L1--M1--L2--M2--X1
        3: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'X1': 'X'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'),
             ('M2.b', 'X1.a')]),
        # L1--M1--L2--M2--L3
        4: Assembly(
            {'L1': 'L', 'M1': 'M', 'L2': 'L', 'M2': 'M', 'L3': 'L'}, 
            [('L1.b', 'M1.a'), ('M1.b', 'L2.a'), ('L2.b', 'M2.a'), 
             ('M2.b', 'L3.a')]),
        }

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        input_path = os.path.join(td, 'input.yaml')
        output_path = os.path.join(td, 'output.yaml')
        wip_dir = os.path.join(td, 'wip')

        with open(input_path, 'w') as f:
            yaml.dump(INPUT_DATA, f)
        
        result = runner.invoke(
            main,
            [
                'enumerate-assemblies',
                str(input_path), str(output_path), 
                '--wip-dir', str(wip_dir), 
                '--overwrite', '--verbose']
        )

        assert result.exit_code == 0
        assert os.path.exists(output_path)

        with open(output_path, 'r') as f:
            actual_output = yaml.safe_load(f)

        for key, assembly in EXPECTED.items():
            assert is_isomorphic(
                actual_output[key], assembly, 
                INPUT_DATA['component_kinds'])  # type: ignore


if __name__ == '__main__':
    pytest.main(['-v', __file__])



import pytest

from nasap_net import InterReaction, IntraReaction


def test_reactants():
    intra_reaction = IntraReaction(
        init_assem_id=0,
        product_assem_id=1,
        leaving_assem_id=None,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=1
    )
    inter_reaction = InterReaction(
        init_assem_id=0,
        entering_assem_id=1,
        product_assem_id=2,
        leaving_assem_id=None,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=1
    )

    assert intra_reaction.reactants == [0]
    assert inter_reaction.reactants == [0, 1]


def test_products_of_intra():
    intra_with_leaving = IntraReaction(
        init_assem_id=0,
        product_assem_id=1,
        leaving_assem_id=2,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=1
    )
    intra_without_leaving = IntraReaction(
        init_assem_id=0,
        product_assem_id=1,
        leaving_assem_id=None,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=1
    )

    assert intra_with_leaving.products == [1, 2]
    assert intra_without_leaving.products == [1]


def test_products_of_inter():
    inter_with_leaving = InterReaction(
        init_assem_id=0,
        entering_assem_id=1,
        product_assem_id=2,
        leaving_assem_id=3,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=1
    )
    inter_without_leaving = InterReaction(
        init_assem_id=0,
        entering_assem_id=1,
        product_assem_id=2,
        leaving_assem_id=None,
        metal_bs='M0.a',
        leaving_bs='X0.a',
        entering_bs='L0.a',
        duplicate_count=1
    )

    assert inter_with_leaving.products == [2, 3]
    assert inter_without_leaving.products == [2]


if __name__ == "__main__":
    pytest.main([__file__, '-v'])



from collections.abc import Iterator
from io import StringIO

import yaml


def split_yaml_documents(yaml_text: str) -> Iterator[str]:
    """
    Safely split YAML text into multiple documents.

    Uses PyYAML's compose_all() to parse the text document-by-document,
    then re-serializes each document and returns it as a string.

    Parameters
    ----------
    yaml_text : str
        YAML string containing multiple documents separated by '---'.

    Yields
    ------
    str
        Each document as an independent YAML string.

    Raises
    ------
    yaml.YAMLError
        If the YAML syntax is invalid.

    Example
    -------
    >>> text = '''
    ... a: 1
    ... ---
    ... b: 2
    ... '''
    >>> for doc in split_yaml_documents(text):
    ...     print('---')
    ...     print(doc.strip())
    ---
    a: 1
    ---
    b: 2
    """
    try:
        docs = yaml.compose_all(yaml_text)
    except yaml.YAMLError as e:
        raise yaml.YAMLError(f"Failed to parse YAML: {e}") from e

    for node in docs:
        buffer = StringIO()
        yaml.serialize(node, stream=buffer)
        yield buffer.getvalue()



from math import inf

import networkx as nx

from nasap_net import Assembly


def find_shortest_path(
        assembly: Assembly,
        comp_id1: str, comp_id2: str,
        ) -> list | None:
    """Find one of the shortest paths between two components in the assembly.

    Parameters
    ----------
    assembly : Assembly
        The assembly object.
    comp_id1 : str
        The component ID of the first component.
    comp_id2 : str
        The component ID of the second component.

    Returns
    -------
    list | None
        The list of component IDs in the shortest path between the two components.
        If no path exists, return None.
    """
    if comp_id1 not in assembly.component_ids:
        raise ValueError(f"Component {comp_id1} is not in the assembly")
    if comp_id2 not in assembly.component_ids:
        raise ValueError(f"Component {comp_id2} is not in the assembly")
    g = assembly.rough_g_snapshot
    try:
        path = nx.shortest_path(g, comp_id1, comp_id2)
        return path
    except nx.NetworkXNoPath:
        return None



from collections.abc import Iterable, Mapping

from .symmetry_application import apply_symmetry_operation

__all__ = ['normalize_bondset_under_sym_ops']


def normalize_bondset_under_sym_ops(
        bondset: Iterable[int],
        sym_ops: Mapping[str, Mapping[int, int]] | None = None
        ) -> set[int]:
    """Find a representative of an assembly under symmetry operations.

    There can be multiple symmetry-equivalent assemblies for a given
    assembly. This function finds the smallest one among them as a
    representative.

    # TODO: Add an example.
    """
    bondset = set(bondset)
    if sym_ops is None:
        return bondset
    for sym_op in sym_ops.values():
        transformed_assem = apply_symmetry_operation(bondset, sym_op)
        if sorted(transformed_assem) < sorted(bondset):
            bondset = transformed_assem
    return bondset



from collections.abc import Iterable, Mapping

__all__ = ['compare_mapping_iterables']


def compare_mapping_iterables(
        iterable1: Iterable[Mapping], iterable2: Iterable[Mapping]
        ) -> bool:
    """
    Compare two iterables of mappings.

    Args:
        iterable1: First iterable of mappings.
        iterable2: Second iterable of mappings.

    Returns:
        bool: True if both iterables contain the same mappings, False otherwise.
    """
    set1 = set(frozenset(d.items()) for d in iterable1)
    set2 = set(frozenset(d.items()) for d in iterable2)
    return set1 == set2



from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from nasap_net.types import ID, SupportsDunderLt
from .binding_site import BindingSite


@dataclass(frozen=True, init=False, order=True)
class Bond(Iterable, SupportsDunderLt):
    """A bond between two binding sites on two components."""
    sites: tuple[BindingSite, BindingSite]

    def __init__(self, comp_id1: ID, site1: ID, comp_id2: ID, site2: ID):
        if comp_id1 == comp_id2:
            raise ValueError("Components in a bond must be different.")
        comp_and_site1 = BindingSite(component_id=comp_id1, site_id=site1)
        comp_and_site2 = BindingSite(component_id=comp_id2, site_id=site2)
        object.__setattr__(
            self, 'sites',
            tuple(sorted((comp_and_site1, comp_and_site2))))  # type:ignore

    def __iter__(self) -> Iterator[BindingSite]:
        return iter(self.sites)

    @property
    def component_ids(self) -> tuple[ID, ID]:
        """Return the component IDs involved in the bond."""
        return self.sites[0].component_id, self.sites[1].component_id

    @classmethod
    def from_sites(cls, site1: BindingSite, site2: BindingSite) -> 'Bond':
        """Create a Bond from two BindingSite instances."""
        return cls(
            comp_id1=site1.component_id,
            comp_id2=site2.component_id,
            site1=site1.site_id,
            site2=site2.site_id
            )



from collections.abc import Iterable
from dataclasses import asdict
from typing import TypeAlias

import pandas as pd

from nasap_net import InterReaction, IntraReaction

Reaction: TypeAlias = IntraReaction | InterReaction


def reactions_to_df(reactions: Iterable[Reaction]) -> pd.DataFrame:
    reactions = list(reactions)
    df = pd.DataFrame(
        [reaction_to_dict(reaction) for reaction in reactions],
        columns=[
            'init_assem_id', 'entering_assem_id', 
            'product_assem_id', 'leaving_assem_id', 
            'metal_bs', 'leaving_bs', 'entering_bs', 
            'duplicate_count', 
        ]
    )
    if isinstance(next(iter(reactions)).init_assem_id, int):
        return df.astype({
            'init_assem_id': int, 'entering_assem_id': pd.Int64Dtype(), 
            'product_assem_id': int, 'leaving_assem_id': pd.Int64Dtype()
        })
    return df


def reaction_to_dict(reaction: Reaction) -> dict:
    d = asdict(reaction)
    if isinstance(reaction, IntraReaction):
        d['entering_assem_id'] = None
    return d



from typing import Any

import yaml

import nasap_net as rx
from nasap_net.utils import FrozenUnorderedPair

from .validations import (validate_name_of_aux_type,
                          validate_name_of_binding_site)

__all__ = ['AuxEdge']


class AuxEdge(yaml.YAMLObject):
    """An auxiliary edge between two binding sites. (Immutable)"""
    yaml_loader = yaml.SafeLoader
    yaml_dumper = yaml.Dumper
    yaml_tag = '!AuxEdge'
    yaml_flow_style = None

    def __init__(
            self, local_binding_site1: str, local_binding_site2: str, 
            aux_kind: str):
        """Initialize an auxiliary edge.

        Note that the order of the binding sites does not matter,
        i.e., AuxEdge('a', 'b', 'cis') is the same as AuxEdge('b', 'a', 'cis').

        Parameters
        ----------
        local_binding_site1 : str
            The local id of the first binding site.
        local_binding_site2 : str
            The local id of the second binding site.
        aux_kind : str
            The auxiliary type.

        Note
        ----
        The order of the binding sites does not matter.

        Raises
        ------
        ValueError
            If the two binding sites are the same.
        """
        validate_name_of_binding_site(local_binding_site1)
        validate_name_of_binding_site(local_binding_site2)
        if local_binding_site1 == local_binding_site2:
            raise ValueError(
                'The two binding sites should be different.')
        self._binding_sites = FrozenUnorderedPair[str](
            local_binding_site1, local_binding_site2)

        validate_name_of_aux_type(aux_kind)
        self._aux_kind = aux_kind
    
    @property
    def local_binding_site1(self) -> str:
        return self._binding_sites.first
    
    @property
    def local_binding_site2(self) -> str:
        return self._binding_sites.second
    
    @property
    def binding_sites(self) -> FrozenUnorderedPair[str]:
        return self._binding_sites

    @property
    def aux_kind(self) -> str:
        return self._aux_kind

    def __hash__(self) -> int:
        return hash((self.binding_sites, self.aux_kind))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, AuxEdge):
            return False
        return (
            self.binding_sites == other.binding_sites and
            self.aux_kind == other.aux_kind)
    
    def __lt__(self, other: 'AuxEdge') -> bool:
        return (sorted(self.binding_sites), self.aux_kind) < (
            sorted(other.binding_sites), other.aux_kind)

    def __repr__(self) -> str:
        return f'AuxEdge({self.local_binding_site1!r}, {self.local_binding_site2!r}, {self.aux_kind!r})'
    
    @classmethod
    def from_yaml(cls, loader, node):
        data = loader.construct_mapping(node, deep=True)
        return AuxEdge(
            data['binding_sites'][0], data['binding_sites'][1], data['aux_kind'])

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_mapping(
            cls.yaml_tag, {
            'binding_sites': sorted(data.binding_sites),
            'aux_kind': data.aux_kind},
            flow_style=cls.yaml_flow_style)



from collections.abc import Iterator, Mapping

from networkx.algorithms.isomorphism import (GraphMatcher,
                                             categorical_edge_match,
                                             categorical_node_match)

from nasap_net import Assembly, Component

__all__ = ['isomorphisms_iter']


def isomorphisms_iter(
        assem1: Assembly, assem2: Assembly,
        component_structures: Mapping[str, Component]
        ) -> Iterator[dict[str, str]]:
    """Find all isomorphisms between two "g_snapshot" graphs of assemblies.

    Parameters
    ----------
    assem1 : Assembly
        The first assembly.
    assem2 : Assembly
        The second assembly.
    component_structures : Mapping[str, ComponentStructure]
        A mapping from component kind to its structure.

    Yields
    ------
    dict[str, str]
        A mapping from node (core or bindsite, e.g., 'M1.core', 'M1.a')
        in the first assembly to node in the second assembly.
    """
    g1 = assem1.g_snapshot(component_structures)
    g2 = assem2.g_snapshot(component_structures)

    node_match = categorical_node_match('component_kind', None)
    edge_match = categorical_edge_match('aux_kind', None)
    GM = GraphMatcher(
        g1, g2, node_match=node_match, edge_match=edge_match)
    yield from GM.isomorphisms_iter()



import os

import pytest
import yaml
from click.testing import CliRunner

from nasap_net import Assembly, Component
from nasap_net.cli.commands import run_explore_reactions_pipeline


@pytest.fixture
def components_data():
    return {
        'component_kinds': {
        'M': Component(['a', 'b']),
        'L': Component(['a', 'b']),
        'X': Component(['a']),
        }
    }


@pytest.fixture
def assemblies():
    return {
        # MX2: X0(a)-(a)M0(b)-(a)X1
        0: Assembly(
            {'M0': 'M', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        ),
        1: Assembly({'L0': 'L'}),  # L: (a)L0(b)
        2: Assembly({'X0': 'X'}),  # X: (a)X0
        # MLX: (a)L0(b)-(a)M0(b)-(a)X0
        3: Assembly(
            {'M0': 'M', 'L0': 'L', 'X0': 'X'},
            [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
        ),
        # ML2: (a)L0(b)-(a)M0(b)-(a)L1(b)
        4: Assembly(
            {'M0': 'M', 'L0': 'L', 'L1': 'L'},
            [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
        ),
        # M2L2X: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)
        5: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'X0': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'L1.a')]
        ),
        # M2LX2: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)X1
        6: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'X1.a')]
        ),
        # M2L2-ring: //-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)-//
        7: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L'},
            [('M0.b', 'L0.a'), ('L0.b', 'M1.a'), ('M1.b', 'L1.a'),
             ('L1.b', 'M0.a')]
        ),
    }


@pytest.fixture
def config():
    return {
        'mle_kinds': {
            'metal': 'M',
            'leaving': 'X',
            'entering': 'L',
        }
    }


@pytest.fixture
def expected_csv():
    return """\
init_assem_id,entering_assem_id,product_assem_id,leaving_assem_id,metal_bs,leaving_bs,entering_bs,duplicate_count
5,,7,2,M0.a,X0.a,L1.b,1
0,1,3,2,M0.a,X0.a,L0.a,4
0,3,6,2,M0.a,X0.a,L0.a,2
0,4,5,2,M0.a,X0.a,L0.a,4
3,1,4,2,M0.b,X0.a,L0.a,2
3,3,5,2,M0.b,X0.a,L0.a,2
6,1,5,2,M0.a,X0.a,L0.a,4
"""


def test_run_enum_assemblies_pipeline(
        tmp_path, assemblies, components_data, config, expected_csv):
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        assemblies_path = os.path.join(td, 'assemblies.yaml')
        components_path = os.path.join(td, 'components.yaml')
        config_path = os.path.join(td, 'config.yaml')
        output_path = os.path.join(td, 'output.csv')

        with open(assemblies_path, 'w') as f:
            yaml.dump(assemblies, f)
        with open(components_path, 'w') as f:
            yaml.dump(components_data, f)
        with open(config_path, 'w') as f:
            yaml.dump(config, f)

        result = runner.invoke(
            run_explore_reactions_pipeline, [
                str(assemblies_path), str(components_path),
                str(config_path), str(output_path),
                '--overwrite', '--verbose'])

        assert result.exit_code == 0

        with open(output_path, 'r') as f:
            output_data = f.read()
        assert output_data == expected_csv


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from collections.abc import Iterable
from typing import TypeVar

__all__ = ['sort_bondsets']


T = TypeVar('T', bound=Iterable[str])


def sort_bondsets(bondsets: Iterable[T]) -> list[T]:
    """Sort bondsets by the number of bonds and the sorted bond IDs.

    The number of bonds is the primary key, and the sorted bond IDs 
    is the secondary key.

    The order of the equal elements is preserved.
    """
    return sorted(
        bondsets,
        key=lambda bondset: (len(list(bondset)), sorted(bondset))
        )


def sort_bondsets_and_bonds(bondsets: Iterable[Iterable[str]]) -> list[list[str]]:
    """Sort bondsets and bonds in the bondsets."""
    return sort_bondsets(
        (sorted(bondset) for bondset in bondsets)
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import ClassVar, Generic, Mapping, TypeVar

from .assembly import Assembly
from .rich_reaction import InterReactionRich, IntraReactionRich, \
    RichReactionBase


class InterOrIntra(Enum):
    INTER = auto()
    INTRA = auto()


R_co = TypeVar("R_co", bound=RichReactionBase, covariant=True)


class ReactionBase(ABC, Generic[R_co]):
    @property
    @abstractmethod
    def inter_or_intra(self) -> InterOrIntra:
        """Whether the reaction is intra- or inter-molecular."""
        pass

    @property
    @abstractmethod
    def reactants(self) -> list[int]:
        """List of reactant assembly IDs."""
        pass

    @property
    @abstractmethod
    def products(self) -> list[int]:
        """List of product assembly IDs."""
        pass

    @abstractmethod
    def to_rich_reaction(self, id_to_assembly: Mapping[int, Assembly]) -> R_co:
        """Convert to a rich reaction by embedding assemblies."""
        pass


@dataclass
class InterReaction(ReactionBase[InterReactionRich]):
    init_assem_id: int
    entering_assem_id: int
    product_assem_id: int
    leaving_assem_id: int | None
    metal_bs: str
    leaving_bs: str
    entering_bs: str
    duplicate_count: int

    @property
    def inter_or_intra(self) -> InterOrIntra:
        return InterOrIntra.INTER

    @property
    def reactants(self) -> list[int]:
        return [self.init_assem_id, self.entering_assem_id]

    @property
    def products(self) -> list[int]:
        if self.leaving_assem_id is None:
            return [self.product_assem_id]
        return [self.product_assem_id, self.leaving_assem_id]

    def to_rich_reaction(
            self, id_to_assembly: Mapping[int, Assembly]
            ) -> InterReactionRich:
        return InterReactionRich.from_reaction(self, id_to_assembly)


@dataclass
class IntraReaction(ReactionBase[IntraReactionRich]):
    init_assem_id: int
    entering_assem_id: ClassVar[None] = None
    product_assem_id: int
    leaving_assem_id: int | None
    metal_bs: str
    leaving_bs: str
    entering_bs: str
    duplicate_count: int

    @property
    def inter_or_intra(self) -> InterOrIntra:
        return InterOrIntra.INTRA

    @property
    def reactants(self) -> list[int]:
        return [self.init_assem_id]

    @property
    def products(self) -> list[int]:
        if self.leaving_assem_id is None:
            return [self.product_assem_id]
        return [self.product_assem_id, self.leaving_assem_id]

    def to_rich_reaction(
            self, id_to_assembly: Mapping[int, Assembly]
            ) -> IntraReactionRich:
        return IntraReactionRich.from_reaction(self, id_to_assembly)



import click

from nasap_net.cli.commands import (run_concat_assembly_lists_pipeline,
                                    run_enum_assemblies_pipeline,
                                    run_explore_reactions_pipeline)


@click.group()
@click.version_option(prog_name='nasap_net')
def main():
    """NASAPNet CLI"""
    pass

main.add_command(run_concat_assembly_lists_pipeline)
main.add_command(run_enum_assemblies_pipeline)
main.add_command(run_explore_reactions_pipeline)



import pytest
import yaml

from nasap_net.pipelines.lib import read_file


def test_basic(tmp_path):
    input_path = tmp_path / "input.yaml"
    data = {"key": "value"}
    with open(input_path, 'w') as f:
        yaml.dump(data, f)
    
    result = read_file(input_path)
    
    assert result == data


def test_nonexistent():
    input_path = "non_existent_file.yaml"
    with pytest.raises(FileNotFoundError):
        read_file(input_path)


def test_invalid_yaml(tmp_path):
    input_path = tmp_path / "invalid.yaml"
    with open(input_path, 'w') as f:
        f.write("key: value: bad_yaml")
    
    with pytest.raises(yaml.YAMLError):
        read_file(input_path)


if __name__ == '__main__':
    pytest.main(['-v', __file__])



import click

from nasap_net.pipelines import (concatenate_assemblies_pipeline,
                                 enumerate_assemblies_pipeline,
                                 explore_reactions_pipeline)


@click.command('enumerate-assemblies')
@click.argument('input', type=click.Path(exists=True))
@click.argument('output', type=click.Path())
@click.option(
    '--wip-dir', '-w', type=click.Path(), 
    help='Directory to store intermediate files.')
@click.option(
    '--overwrite', '-o',
    is_flag=True, help='Overwrite output file if it exists.')
@click.option(
    '--verbose', '-v',
    is_flag=True, help='Print verbose output.')
def run_enum_assemblies_pipeline(input, output, wip_dir, overwrite, verbose):
    """Enumerates assemblies.
    
    \b
    Parameters
    ----------
    - INPUT: Path to input file.
    - OUTPUT: Path to output file.

    \b
    Options
    -------
    --wip-dir, -w: Directory to store intermediate files.
    --overwrite, -o: Overwrite output file if it exists.
    --verbose, -v: Print verbose output
    """
    enumerate_assemblies_pipeline(
        input, output,
        wip_dir=wip_dir, overwrite=overwrite, verbose=verbose)


@click.command('concat-assembly-lists')
@click.argument('assemblies', type=click.Path(exists=True), nargs=-1)
@click.argument('component_kinds', type=click.Path(exists=True))
@click.argument('output', type=click.Path())
@click.option(
    '--already-unique-within-files', '-u', is_flag=True,
    help='Whether the assemblies in each file are already unique. If used, '
    'the isomorphism checks are skipped for the assemblies within each file.')
@click.option(
    '--skip-isomorphism-checks', '-i', is_flag=True,
    help='Whether to skip all isomorphism checks. If True, the assemblies '
    'are concatenated without checking for isomorphism. The resulting list '
    'may contain duplicate assemblies.')
@click.option(
    '--start', '-s', type=int, default=0,
    help='Starting index for the reindexing of the assemblies.')
@click.option(
    '--overwrite', '-o', is_flag=True,
    help='Overwrite output file if it exists.')
@click.option(
    '--verbose', '-v', is_flag=True,
    help='Print verbose output.')
def run_concat_assembly_lists_pipeline(
        assemblies, component_kinds, output, already_unique_within_files, 
        skip_isomorphism_checks, start, overwrite, verbose):
    """Concatenates assembly lists.

    \b
    Parameters
    ----------
    - ASSEMBLIES: Paths to input files of assemblies.
    - COMPONENT_KINDS: Path to input file of component kinds.
    - OUTPUT: Path to output file.

    \b
    Options
    -------
    --already-unique-within-files, -u: Whether the assemblies in each file are already unique. If used, the isomorphism checks are skipped for the assemblies within each file.
    --skip-isomorphism-checks, -i: Whether to skip all isomorphism checks. 
    If True, the assemblies are concatenated without checking for 
    isomorphism. The resulting list may contain duplicate assemblies.
    --start, -s: Starting index for the reindexing of the assemblies.
    --overwrite, -o: Overwrite output file if it exists.
    --verbose, -v: Print verbose output.
    """
    concatenate_assemblies_pipeline(
        assemblies, component_kinds, output,
        already_unique_within_files=already_unique_within_files,
        skip_isomorphism_checks=skip_isomorphism_checks,
        start=start, overwrite=overwrite, verbose=verbose)


@click.command('explore-reactions')
@click.argument('assemblies', type=click.Path(exists=True))
@click.argument('component_kinds', type=click.Path(exists=True))
@click.argument('config', type=click.Path(exists=True))
@click.argument('output', type=click.Path())
@click.option(
    '--overwrite', '-o', is_flag=True,
    help='Overwrite output file if it exists.')
@click.option(
    '--verbose', '-v', is_flag=True,
    help='Print verbose output.')
def run_explore_reactions_pipeline(
        assemblies, component_kinds, config, output, overwrite, verbose):
    """Explores reactions in assemblies.
    \b
    Parameters
    ----------
    - ASSEMBLIES: Path to input file of assemblies.
    - COMPONENT_KINDS: Path to input file of component kinds.
    - CONFIG: Path to input file of configuration.
    - OUTPUT: Path to output file.
    \b
    Options
    -------
    --overwrite, -o: Overwrite output file if it exists.
    --verbose, -v: Print verbose output.
    """
    explore_reactions_pipeline(
        assemblies, component_kinds, config, output,
        overwrite=overwrite, verbose=verbose)



import pytest

from nasap_net.pipelines.bondset_enumeration import validate_sym_ops


def test_valid():
    bonds = [1, 2, 3]
    sym_ops = {
        'C3': {1: 2, 2: 3, 3: 1},
        'C3^2': {1: 3, 2: 1, 3: 2}
    }
    validate_sym_ops(sym_ops, bonds)


def test_invalid_keys():
    bonds = [1, 2, 3]
    sym_ops = {
        'C3': {1: 2, 2: 3, 4: 1},  # Missing key 3
    }
    with pytest.raises(
            ValueError, 
            match='Keys in the operation mapping must contain all bonds'):
        validate_sym_ops(sym_ops, bonds)


def test_invalid_values():
    bonds = [1, 2, 3]
    sym_ops = {
        'C3': {1: 2, 2: 3, 3: 2},  # Missing value 1
    }
    with pytest.raises(
            ValueError, 
            match='Values in the operation mapping must contain all bonds'):
        validate_sym_ops(sym_ops, bonds)


if __name__ == '__main__':
    pytest.main(['-v', __file__])



import pytest
import yaml

from nasap_net import Assembly, Component, is_isomorphic
from nasap_net.pipelines import cap_bindsites_pipeline


@pytest.fixture
def assemblies_data():
    return {
        0: Assembly({'M1': 'M'}),  # M1
        1: Assembly({'L1': 'L', 'M1': 'M'}, [('L1.b', 'M1.a')]),  # L1--M1
        2: Assembly(  # M1--L1--M2
            {'M1': 'M', 'L1': 'L', 'M2': 'M'},
            [('M1.b', 'L1.a'), ('L1.b', 'M2.a')]),
    }


@pytest.fixture
def components_data():
    return {'component_kinds': {
        'L': Component(['a', 'b']),
        'M': Component(['a', 'b']),
        'X': Component(['a']),
    }}


def test_add_X_on_M(
        tmp_path, assemblies_data, components_data):
    assemblies_path = tmp_path / 'assemblies.yaml'
    components_path = tmp_path / 'components.yaml'
    config_path = tmp_path / 'config.yaml'
    output_path = tmp_path / 'output.yaml'

    CONFIG_DATA = {'capping_config': {
        'target_component_kind': 'M',
        'capping_component_kind': 'X',
        'capping_binding_site': 'a'
    }}

    EXPECTED_CAPPED_ASSEMBLIES = {
        0: Assembly(  # X1--M1--X2
            {'M1': 'M', 'X1': 'X', 'X2': 'X'}, 
            [('X1.a', 'M1.a'), ('X2.a', 'M1.b')]),
        1: Assembly(  # L1--M1--X1
            {'L1': 'L', 'M1': 'M', 'X1': 'X'}, 
            [('L1.b', 'M1.a'), ('X1.a', 'M1.b')]),
        2: Assembly(  # X1--M1--L1--M2--X2
            {'M1': 'M', 'L1': 'L', 'M2': 'M', 'X1': 'X', 'X2': 'X'}, 
            [('X1.a', 'M1.a'), ('M1.b', 'L1.a'), ('L1.b', 'M2.a'), 
             ('M2.b', 'X2.a')]),
        }

    with open(assemblies_path, 'w') as f:
        yaml.dump(assemblies_data, f)
    
    with open(components_path, 'w') as f:
        yaml.dump(components_data, f)
    
    with open(config_path, 'w') as f:
        yaml.dump(CONFIG_DATA, f)

    # Run the pipeline
    cap_bindsites_pipeline(
        assemblies_path=str(assemblies_path),
        components_path=str(components_path),
        config_path=str(config_path),
        output_path=str(output_path),
    )

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data.keys() == EXPECTED_CAPPED_ASSEMBLIES.keys()
    for assembly_id, expected_assembly in EXPECTED_CAPPED_ASSEMBLIES.items():
        assert is_isomorphic(
            output_data[assembly_id], expected_assembly,
            components_data['component_kinds'])


def test_add_L_on_M(
        tmp_path, assemblies_data, components_data):
    assemblies_path = tmp_path / 'assemblies.yaml'
    components_path = tmp_path / 'components.yaml'
    config_path = tmp_path / 'config.yaml'
    output_path = tmp_path / 'output.yaml'

    CONFIG_DATA = {
        'capping_config': {
        'target_component_kind': 'M',
        'capping_component_kind': 'L',
        'capping_binding_site': 'a'
    }}

    EXPECTED_CAPPED_ASSEMBLIES = {
        0: Assembly(  # L1--M1--L2
            {'M1': 'M', 'L1': 'L', 'L2': 'L'}, 
            [('L1.a', 'M1.a'), ('L2.a', 'M1.b')]),
        1: Assembly(  # L1--M1--L2
            {'L1': 'L', 'M1': 'M', 'L2': 'L'}, 
            [('L1.b', 'M1.a'), ('L2.a', 'M1.b')]),
        2: Assembly(  # L1--M1--L2--M2--L3
            {'M1': 'M', 'L1': 'L', 'L2': 'L', 'M2': 'M', 'L3': 'L'},
            [('L1.a', 'M1.a'), ('L2.a', 'M1.b'), ('L2.b', 'M2.a'),
             ('L3.a', 'M2.b')]),
        }

    with open(assemblies_path, 'w') as f:
        yaml.dump(assemblies_data, f)
    
    with open(components_path, 'w') as f:
        yaml.dump(components_data, f)
    
    with open(config_path, 'w') as f:
        yaml.dump(CONFIG_DATA, f)

    # Run the pipeline
    cap_bindsites_pipeline(
        assemblies_path=str(assemblies_path),
        components_path=str(components_path),
        config_path=str(config_path),
        output_path=str(output_path),
    )

    with open(output_path, 'r') as f:
        output_data = yaml.safe_load(f)

    assert output_data.keys() == EXPECTED_CAPPED_ASSEMBLIES.keys()
    for assembly_id, expected_assembly in EXPECTED_CAPPED_ASSEMBLIES.items():
        assert is_isomorphic(
            output_data[assembly_id], expected_assembly,
            components_data['component_kinds'])
        

if __name__ == '__main__':
    pytest.main(['-v', __file__])



from nasap_net.models import Assembly, AuxEdge, Bond, Component
from nasap_net.yaml import load


def test_load():
    yaml_str = """M: !Component
  kind: M
  sites: [0, 1]
M(aux): !Component
  kind: M(aux)
  sites: [0, 1, 2]
  aux_edges:
  - sites: [0, 1]
  - sites: [0, 2]
    kind: cis
X: !Component
  kind: X
  sites: [0]
---
M(aux)X3: !LightAssembly
  components: {M0: M(aux), X0: X, X1: X, X2: X}
  bonds:
  - [M0, 0, X0, 0]
  - [M0, 1, X1, 0]
  - [M0, 2, X2, 0]
MX2: !LightAssembly
  components: {M0: M, X0: X, X1: X}
  bonds:
  - [M0, 0, X0, 0]
  - [M0, 1, X1, 0]
  id: MX2
free_X: !LightAssembly
  components: {X0: X}
  bonds: []
"""

    loaded = load(yaml_str)

    M = Component(kind='M', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    M_aux = Component(
        kind='M(aux)', sites=[0, 1, 2],
        aux_edges=[AuxEdge(0, 1), AuxEdge(0, 2, kind='cis')])

    assert loaded == {
        'free_X': Assembly(components={'X0': X}, bonds=[]),
        # MX2: X0(0)-(0)M0(1)-(0)X1
        'MX2': Assembly(
            id_='MX2',
            components={'X0': X, 'M0': M, 'X1': X},
            bonds=[Bond('X0', 0, 'M0', 0), Bond('M0', 1, 'X1', 0)]),
        'M(aux)X3': Assembly(
            components={'M0': M_aux, 'X0': X, 'X1': X, 'X2': X},
            bonds=[
                Bond('M0', 0, 'X0', 0), Bond('M0', 1, 'X1', 0),
                Bond('M0', 2, 'X2', 0)
            ]
        ),
    }



import pytest

from nasap_net import Assembly, Component, assemblies_equal, perform_intra_exchange


def test_intra_exchange() -> None:
    # X1(a)--(a)M1(b)--(a)L1(b)
    MLX = Assembly(
        {'M1': 'M', 'X1': 'X', 'L1': 'L'},
        [('X1.a', 'M1.a'), ('M1.b', 'L1.a')])
    # Expected product: ML ring
    ML_RING = Assembly(
        {'M1': 'M', 'L1': 'L'},
        [('M1.b', 'L1.a'), ('L1.b', 'M1.a')])
    # Expected leaving: X
    X = Assembly({'X1': 'X'}, [])

    COMPONENT_KINDS = {
        'M': Component({'a', 'b'}),
        'L': Component({'a', 'b'}),
        'X': Component({'a'})}
    
    product, leaving = perform_intra_exchange(
        MLX, 'M1.a', 'X1.a', 'L1.b')

    assert assemblies_equal(product, ML_RING, COMPONENT_KINDS)
    assert leaving is not None
    assert assemblies_equal(leaving, X, COMPONENT_KINDS)


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Mapping, Sequence
from itertools import chain

from nasap_net import Assembly, Component
from nasap_net.algorithms.isomorphism import isomorphisms_iter


def are_equivalent_binding_site_lists(
        assembly: Assembly,
        binding_site_list1: Sequence[str],
        binding_site_list2: Sequence[str],
        component_structures: Mapping[str, Component],
        ) -> bool:
    """
    Check if the pair of binding site lists are equivalent.

    "Equivalent" means that there is at least one isomorphism 
    which maps each binding site in the first list to a binding site
    in the second list.
    The order of the binding sites in the list DOES matter.

    Parameters
    ----------
    assembly : Assembly
        The assembly to which the binding sites belong.
    binding_site_list1 : Sequence[str]
        The first list of binding sites.
    binding_site_list2 : Sequence[str]
        The second list of binding sites.
    component_structures : Mapping[str, Component]
        The component structures of the assembly.

    Returns
    -------
    bool
        True if the binding site lists are equivalent, False otherwise.
    """
    # Input validation
    if len(binding_site_list1) != len(binding_site_list2):
        raise ValueError(
            "The length of the binding site lists must be the same.")
    if not binding_site_list1:
        raise ValueError("The binding site lists must not be empty.")
    
    # Check if the binding sites are in the assembly
    for bs in chain(binding_site_list1, binding_site_list2):
        if bs not in assembly.get_all_bindsites(component_structures):
            raise ValueError(f"Binding site {bs} not found in assembly.")

    # ---
    # Main logic
    for isom in isomorphisms_iter(
            assembly, assembly, component_structures):
        for (bs1, bs2) in zip(binding_site_list1, binding_site_list2):
            if isom[bs1] != bs2:
                break
        else:
            return True
    return False



import random
from copy import deepcopy
from typing import TypeAlias

import pytest

from nasap_net import Assembly, Component, InterReaction, IntraReaction
from nasap_net.algorithms import are_equivalent_reaction_sets
from nasap_net.algorithms.reaction_comparison.multiple_reaction_comparison import \
    _group_reactions

Reaction: TypeAlias = IntraReaction | InterReaction


@pytest.fixture
def component_structures():
    return {
        'L': Component(['a', 'b']),
        'M': Component(['a', 'b']),
        'X': Component(['a']),
    }


@pytest.fixture
def id_to_assembly():
    return {
        # MX2: X0(a)-(a)M0(b)-(a)X1
        0: Assembly(
            {'M0': 'M', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        ),
        1: Assembly({'L0': 'L'}),  # L: (a)L0(b)
        2: Assembly({'X0': 'X'}),  # X: (a)X0
        # MLX: (a)L0(b)-(a)M0(b)-(a)X0
        3: Assembly(
            {'M0': 'M', 'L0': 'L', 'X0': 'X'},
            [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
        ),
        # ML2: (a)L0(b)-(a)M0(b)-(a)L1(b)
        4: Assembly(
            {'M0': 'M', 'L0': 'L', 'L1': 'L'},
            [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
        ),
        # M2L2X: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)
        5: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'X0': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'L1.a')]
        ),
        # M2LX2: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)X1
        6: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'X1.a')]
        ),
        # M2L2-ring: //-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)-//
        7: Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L'},
            [('M0.b', 'L0.a'), ('L0.b', 'M1.a'), ('M1.b', 'L1.a'),
             ('L1.b', 'M0.a')]
        ),
    }


def test_list_with_single_reaction(component_structures, id_to_assembly):
    # MX2 + L -> MLX + X
    reaction1 = InterReaction(
        init_assem_id=0, entering_assem_id=1,
        product_assem_id=3, leaving_assem_id=2,
        metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
        duplicate_count=4,
    )
    # Equivalent reaction
    reaction2 = InterReaction(
        init_assem_id=0, entering_assem_id=1,
        product_assem_id=3, leaving_assem_id=2,
        metal_bs='M0.b',  # Different metal binding site
        leaving_bs='X1.a',  # Different leaving binding site
        entering_bs='L0.a',
        duplicate_count=4,
    )

    assert are_equivalent_reaction_sets(
        [reaction1], [reaction2], id_to_assembly, component_structures)


def test_list_with_multiple_reactions(component_structures, id_to_assembly):
    # MX2 + L -> MLX + X
    reaction1 = InterReaction(
        init_assem_id=0, entering_assem_id=1,
        product_assem_id=3, leaving_assem_id=2,
        metal_bs='M0.a', leaving_bs='X0.a',
        entering_bs='L0.a', duplicate_count=4,
    )
    # Equivalent reaction
    reaction2 = InterReaction(
        init_assem_id=0, entering_assem_id=1,
        product_assem_id=3, leaving_assem_id=2,
        metal_bs='M0.b',  # Different metal binding site
        leaving_bs='X1.a',  # Different leaving binding site
        entering_bs='L0.a',
        duplicate_count=4,
    )

    # Add an additional assembly M2L3
    # M2L3: (a)L0(b)-(a)M0(b)-(a)L1(b)-(a)M1(b)-(a)L2(b)
    id_to_assembly[8] = Assembly(
        {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'L2': 'L'},
        [('L0.b', 'M0.a'), ('M0.b', 'L1.a'), ('L1.b', 'M1.a'),
         ('M1.b', 'L2.a')]
    )

    # M2L3 is needed for "the equivalent intra-molecular reactions
    # with different binding sites" described below.

    # M2L3 -> M2L2-ring + L
    reaction3 = IntraReaction(
        init_assem_id=8, product_assem_id=7,
        leaving_assem_id=1, metal_bs='M0.a',
        leaving_bs='L0.b', entering_bs='L2.b', duplicate_count=2,
    )
    # Equivalent reaction
    reaction4 = IntraReaction(
        init_assem_id=8, product_assem_id=7, leaving_assem_id=1,
        metal_bs='M1.b',  # Different metal binding site
        leaving_bs='L2.a',  # Different leaving binding site
        entering_bs='L0.a',
        duplicate_count=2,
    )

    assert are_equivalent_reaction_sets(
        [reaction1, reaction3], [reaction2, reaction4], id_to_assembly,
        component_structures)


def test_comprehensive_reaction_sets(component_structures, id_to_assembly):

    # All "X to L" reactions among the assemblies in id_to_assembly
    # including intra- and inter-molecular reactions:
    reactions1: dict[str, Reaction] = {
        # Intra-molecular reactions
        'M2L2X -> M2L2-ring + X': IntraReaction(
            init_assem_id=5, 
            product_assem_id=7, leaving_assem_id=2, 
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L1.b',
            duplicate_count=1,
        ),
        # Inter-molecular reactions
        'MX2 + L -> MLX + X': InterReaction(
            init_assem_id=0, entering_assem_id=1,
            product_assem_id=3, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
        'MX2 + MLX -> M2LX2 + X': InterReaction(
            init_assem_id=0, entering_assem_id=3,
            product_assem_id=6, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'MX2 + ML2 -> M2L2X + X': InterReaction(
            init_assem_id=0, entering_assem_id=4,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
        'MLX + L -> ML2 + X': InterReaction(
            init_assem_id=3, entering_assem_id=1,
            product_assem_id=4, leaving_assem_id=2,
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'MLX + MLX -> M2L2X + X': InterReaction(
            init_assem_id=3, entering_assem_id=3,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=2,
        ),
        'M2LX2 + L -> M2L2X + X': InterReaction(
            init_assem_id=6, entering_assem_id=1,
            product_assem_id=5, leaving_assem_id=2,
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a',
            duplicate_count=4,
        ),
    }
    
    reactions2 = deepcopy(reactions1)

    # Use different binding sites for some reactions
    reactions2['MX2 + L -> MLX + X'] = InterReaction(
            init_assem_id=0, entering_assem_id=1,
            product_assem_id=3, leaving_assem_id=2,
            # Different metal and leaving binding sites
            metal_bs='M0.b', leaving_bs='X1.a', entering_bs='L0.a',
            duplicate_count=4,
        )
    reactions2['MLX + L -> ML2 + X'] = InterReaction(
            init_assem_id=3, entering_assem_id=1,
            product_assem_id=4, leaving_assem_id=2,
            # Different entering binding site
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.b',
            duplicate_count=2,
        )
    reactions2['M2LX2 + L -> M2L2X + X'] = InterReaction(
            init_assem_id=6, entering_assem_id=1,
            product_assem_id=5, leaving_assem_id=2,
            # Different metal, leaving, and entering binding sites
            metal_bs='M1.b', leaving_bs='X1.a', entering_bs='L0.b',
            duplicate_count=4,
        )
    
    reaction_list1 = list(reactions1.values())
    reaction_list2 = list(reactions2.values())

    # Randomly shuffle the reaction_list2
    random.shuffle(reaction_list2)

    assert are_equivalent_reaction_sets(
        reaction_list1, reaction_list2, id_to_assembly,
        component_structures
    ), "The two sets of reactions should be equivalent."


def test_group_reactions_basic():
    # Two reactions with the same grouping key, one with a different key
    r1 = InterReaction(
        init_assem_id=1, entering_assem_id=2,
        product_assem_id=3, leaving_assem_id=4,
        metal_bs='a', leaving_bs='a', entering_bs='a',
        duplicate_count=1,
    )
    r2 = InterReaction(
        # same assemblies
        init_assem_id=1, entering_assem_id=2,
        product_assem_id=3, leaving_assem_id=4,
        # different binding sites
        metal_bs='b', leaving_bs='b', entering_bs='b',
        # different duplicate count
        duplicate_count=2,
    )
    r3 = InterReaction(
        # different assemblies
        init_assem_id=5, entering_assem_id=6,
        product_assem_id=7, leaving_assem_id=8,
        # same binding sites
        metal_bs='a', leaving_bs='a', entering_bs='a',
        # same duplicate count
        duplicate_count=1,
    )
    grouped = _group_reactions([r1, r2, r3])
    assert len(grouped) == 2
    key1 = (1, 2, 3, 4)
    key3 = (5, 6, 7, 8)
    assert key1 in grouped and key3 in grouped
    assert grouped[key1] == [r1, r2] or grouped[key1] == [r2, r1]
    assert grouped[key3] == [r3]


if __name__ == "__main__":
    pytest.main([__file__, '-v'])



import pytest

from nasap_net import Assembly, InterReactionRich, ReactionClassifier


def test():
    def classification_rule(reaction):
        if (reaction.leaving_kind == 'X'
                and reaction.entering_kind == 'L'):
            return 'X-L exchange'
        else:
            return 'other'
        
    classifier = ReactionClassifier(classification_rule)

    MX = Assembly({'M1': 'M', 'X1': 'X'}, [('M1.a', 'X1.a')])
    L = Assembly({'L1': 'L'})
    ML = Assembly({'M1': 'M', 'L1': 'L'}, [('M1.a', 'L1.a')])
    X = Assembly({'X1': 'X'})
    
    X_to_L = InterReactionRich(
        init_assem=MX, entering_assem=L,
        product_assem=ML, leaving_assem=X,
        metal_bs='M1.a', leaving_bs='X1.a', entering_bs='L1.a',
        duplicate_count=1
        )
    L_to_X = InterReactionRich(
        init_assem=ML, entering_assem=X,
        product_assem=MX, leaving_assem=L,
        metal_bs='M1.a', leaving_bs='L1.a', entering_bs='X1.a',
        duplicate_count=1
        )
    assert classifier.classify(X_to_L) == 'X-L exchange'
    assert classifier.classify(L_to_X) == 'other'


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



from collections.abc import Mapping

from nasap_net import Assembly, Component

from .lib import group_assemblies_by_isomorphism

__all__ = ['exclude_remaining_duplicates']


def exclude_remaining_duplicates(
        id_to_assembly: Mapping[int, Assembly],
        component_structures: Mapping[str, Component],
        ) -> tuple[dict[int, Assembly], dict[int, set[int]]]:
    """Exclude remaining duplicates."""
    unique_id_to_ids = group_assemblies_by_isomorphism(id_to_assembly, component_structures)
    unique_ids = unique_id_to_ids.keys()
    unique_assemblies = {id_: id_to_assembly[id_] for id_ in unique_ids}
    return unique_assemblies, unique_id_to_ids



import os
from collections.abc import Hashable, Mapping
from typing import TypeVar

from nasap_net import Assembly, Component, group_assemblies_by_isomorphism
from nasap_net.pipelines.lib import read_file, write_output

_T = TypeVar('_T', bound=Hashable)



# ============================================================
# Main Process
# ============================================================

def find_unique_assemblies_pipeline(
        assemblies_path: os.PathLike | str,
        components_path: os.PathLike | str,
        output_path: os.PathLike | str,
        *,
        overwrite: bool = False,
        verbose: bool = False
        ) -> None:
    # Input
    id_to_assembly: Mapping[Hashable, Assembly] = read_file(
        assemblies_path, verbose=verbose)
    components: Mapping[str, Component] = read_file(
        components_path, verbose=verbose)['component_kinds']
    
    # Main process
    if verbose:
        print('Enumerating unique assemblies...')
    
    unique_id_to_dup_ids = group_assemblies_by_isomorphism(
        id_to_assembly, components)
    id_to_unique_assembly = {
        unique_id: id_to_assembly[unique_id]
        for unique_id in unique_id_to_dup_ids
    }
    
    if verbose:
        print('Finished enumeration.')
    
    # Output
    write_output(
        output_path, id_to_unique_assembly, 
        overwrite=overwrite, verbose=verbose,
        header='Unique assemblies')



from collections import defaultdict
from collections.abc import Iterator, Mapping, Sequence

from nasap_net import Assembly, Component, calc_graph_hash_of_assembly

from .lib.is_new_check import is_new

__all__ = ['find_unique_assemblies']


def find_unique_assemblies(
        assemblies: Sequence[Assembly],
        component_structures: Mapping[str, Component]
        ) -> Iterator[Assembly]:
    """Get unique assemblies.

    "Unique" here means that the assemblies are not isomorphic to each
    other. See `stepwise_is_isomorphic` for the definition of isomorphism.
    
    Although `exclude_remaining_duplicates` provides a more detailed 
    information like which unique assembly corresponds to which duplicates, 
    this function is more efficient in terms of memory usage if you only 
    need the unique assemblies.

    Parameters
    ----------
    assemblies : Sequence[Assembly]
        A sequence of assemblies.
    component_structures : Mapping[str, Component]
        A mapping from component kinds to component structures.

    Yields
    ------
    tuple[str, Assembly]
        Unique assemblies with their IDs, i.e., (ID, assembly).

    Note
    ----
    From each group of isomorphic assemblies, the first assembly in the
    provided iterable is yielded as a unique assembly.
    """
    hash_to_uniques: dict[str, list[Assembly]] = defaultdict(list)
    for assembly in assemblies:
        hash_ = calc_graph_hash_of_assembly(assembly, component_structures)
        if is_new(
                hash_, assembly, hash_to_uniques, component_structures):
            hash_to_uniques[hash_].append(assembly)
            yield assembly



from collections.abc import Mapping
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D

from nasap_net import Assembly, Component

from .color_map import make_complete_color_map

__all__ = ['draw_3d']


def draw_3d(
        assembly: Assembly,
        component_structures: Mapping[str, Component],
        positions: Mapping[str, tuple[float, float, float]],
        *,
        show: bool = True,
        # Sizes and colors
        core_node_size: int = 1600,
        bindsite_node_size: int = 400,
        component_colors: Mapping[str, str] = {},
        node_alpha: float = .8,
        edge_alpha: float = .8,
        core_to_bindsite_edge_width: float = 8,
        bond_edge_width: float = 2,
        # Layout
        figsize: tuple[float, float] = (6, 6),
        x_lim: tuple[float, float] | None = None,
        y_lim: tuple[float, float] | None = None,
        z_lim: tuple[float, float] | None = None,
        projection_type: Literal['ortho', 'persp'] = 'ortho',
        # Labels
        node_labeling_mode: Literal[
            'component_kind', 'component_id', 'core_and_bindsite_ids',
            None] = 'component_kind',
        node_labeling_args: Mapping = {'font_size': 12},
        label_aux_edges: bool = True,
        aux_edge_labeling_args: Mapping = {'font_size': 12},
        # Others
        show_axes: bool = False,
        show_grid: bool = False,
        ) -> None:
    component_colors = make_complete_color_map(assembly, component_colors)

    fig = plt.figure(figsize=figsize)
    ax: Axes3D = fig.add_subplot(111, projection='3d')
    ax.set_aspect('equal')

    if x_lim is not None:
        ax.set_xlim(x_lim)
    if y_lim is not None:
        ax.set_ylim(y_lim)
    
    g = assembly.g_snapshot(component_structures)

    # ========= Draw the graph =========

    # Draw cores
    cores = list(assembly.iter_all_cores())

    core_fill_colors = [
        component_colors[assembly.get_component_kind_of_core(n)] 
        for n in cores]
    
    core_pos = np.array([positions[n] for n in cores])
    
    ax.scatter(
        *core_pos.T, alpha=node_alpha, s=core_node_size,
        color=core_fill_colors
        )  # type: ignore
    
    # Draw bindsite
    bindsites = list(assembly.get_all_bindsites(component_structures))

    bindsite_edge_colors = [
        component_colors[assembly.get_component_kind_of_bindsite(n)] 
        for n in bindsites]
    
    bindsite_pos = np.array([positions[n] for n in bindsites])
    
    ax.scatter(
        *bindsite_pos.T, alpha=node_alpha, s=bindsite_node_size,
        color='#FFF', 
        edgecolors=bindsite_edge_colors, linewidths=2
        )  # type: ignore
    
    # Draw core-to-bindsite edges
    for comp_id, comp_kind in assembly.comp_id_to_kind.items():
        color = component_colors[comp_kind]
        core = assembly.get_core_of_the_component(comp_id)
        bindsites_of_the_comp = assembly.get_bindsites_of_component(
            comp_id, component_structures)
        
        for bindsite in bindsites_of_the_comp:
            xs = [positions[core][0], positions[bindsite][0]]
            ys = [positions[core][1], positions[bindsite][1]]
            zs = [positions[core][2], positions[bindsite][2]]
            
            ax.plot(
                xs, ys, zs, color=color, alpha=edge_alpha, 
                linewidth=core_to_bindsite_edge_width)
    
    # Draw bonds
    for bindsite1, bindsite2 in assembly.bonds:
        xs = [positions[bindsite1][0], positions[bindsite2][0]]
        ys = [positions[bindsite1][1], positions[bindsite2][1]]
        zs = [positions[bindsite1][2], positions[bindsite2][2]]

        ax.plot(
            xs, ys, zs, color='grey', alpha=edge_alpha, 
            linestyle=(0, (2, 1)), linewidth=bond_edge_width)
        
    # Draw aux edges
    for aux_edge in assembly.iter_aux_edges(component_structures):
        bs1 = aux_edge.bindsite1
        bs2 = aux_edge.bindsite2
        xs = [positions[bs1][0], positions[bs2][0]]
        ys = [positions[bs1][1], positions[bs2][1]]
        zs = [positions[bs1][2], positions[bs2][2]]

        ax.plot(
            xs, ys, zs, color='darkgrey', alpha=edge_alpha, 
            linestyle='dotted')

    # Draw the labels
    labels = {}
    match node_labeling_mode:
        case 'component_kind':
            for comp_id, comp_kind in assembly.comp_id_to_kind.items():
                core = assembly.get_core_of_the_component(comp_id)
                labels[core] = comp_kind
        case 'component_id':
            for comp_id in assembly.component_ids:
                core = assembly.get_core_of_the_component(comp_id)
                labels[core] = comp_id
        case 'core_and_bindsite_ids':
            for core in cores:
                labels[core] = core
            for bindsite in bindsites:
                labels[bindsite] = bindsite

    node_labeling_ax_text_args = {}
    node_labeling_ax_text_args['fontsize'] = node_labeling_args.get('font_size', 12)
    node_labeling_ax_text_args['horizontalalignment'] = node_labeling_args.get(
        'horizontalalignment', 'center')
    node_labeling_ax_text_args['verticalalignment'] = node_labeling_args.get(
        'verticalalignment', 'center')

    if node_labeling_mode is not None:
        for node, label in labels.items():
            x, y, z = positions[node]
            ax.text(x, y, z, label, **node_labeling_ax_text_args)

    aux_labeling_ax_text_args = {}
    aux_labeling_ax_text_args['fontsize'] = aux_edge_labeling_args.get('font_size', 12)
    aux_labeling_ax_text_args['horizontalalignment'] = aux_edge_labeling_args.get(
        'horizontalalignment', 'center')
    aux_labeling_ax_text_args['verticalalignment'] = aux_edge_labeling_args.get(
        'verticalalignment', 'center')
    
    if label_aux_edges:
        for aux_edge in assembly.iter_aux_edges(component_structures):
            bs1 = aux_edge.bindsite1
            bs2 = aux_edge.bindsite2
            aux_kind = aux_edge.aux_type
            x = (positions[bs1][0] + positions[bs2][0]) / 2
            y = (positions[bs1][1] + positions[bs2][1]) / 2
            z = (positions[bs1][2] + positions[bs2][2]) / 2
            ax.text(x, y, z, aux_kind, **aux_labeling_ax_text_args)

    if not show_axes:
        ax.set_axis_off()
    if not show_grid:
        ax.grid(False)
    plt.tight_layout()

    ax.set_xlim(x_lim)
    ax.set_ylim(y_lim)
    ax.set_zlim(z_lim)

    if projection_type == 'ortho':
        ax.set_proj_type('ortho')
    elif projection_type == 'persp':
        ax.set_proj_type('persp')

    if show:
        plt.show()



"""
NASAPNet
========

NASAPNet: Automates the construction of reaction networks as part of the NASAP (Numerical Analysis for Self-Assembly Process) methodology, originally developed by the Hiraoka Group.

See the documentation for details.
"""
# isort: skip_file

# 'isort: split' is used to separate the imports into groups.
# (although the comment 'isort: split' itself is not necessary
# as long as 'isort: skip_file' is used.)

__version__ = "0.1.0"

from .classes import *

# isort: split
from .algorithms import *

# isort: split
from .assembly_drawing import draw_2d
from .assembly_drawing import draw_3d

from .bondset_enumeration import enum_bond_subsets

from .bondset_to_assembly import convert_bondset_to_assembly
from .bondset_to_assembly import convert_bondsets_to_assemblies

from .bindsite_capping import cap_bindsites
from .bindsite_capping import cap_single_bindsite

from .duplicate_exclusion import find_unique_assemblies
from .duplicate_exclusion import group_assemblies_by_isomorphism

from .reaction_exploration import explore_reactions

from .reaction_classification import ReactionClassifier


from dataclasses import dataclass

from nasap_net.types import ID, SupportsDunderLt


@dataclass(frozen=True, order=True)
class BindingSite(SupportsDunderLt):
    """A specific binding site on a specific component."""
    component_id: ID
    site_id: ID



from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import TypeAlias, cast

from nasap_net import Assembly, Component, InterReaction, IntraReaction

from .reaction_comparison import are_equivalent_reactions

Reaction: TypeAlias = IntraReaction | InterReaction


def are_equivalent_reaction_sets(
        reaction_set1: Iterable[Reaction],
        reaction_set2: Iterable[Reaction],
        id_to_assembly: Mapping[int, Assembly],
        component_structures: Mapping[str, Component],
        ) -> bool:
    """Check if the two sets of reactions are equivalent.
    """
    reactions1 = list(reaction_set1)  # e.g. [r1-0, r1-1, r1-2]
    reactions2 = list(reaction_set2)  # e.g. [r2-1, r2-2, r2-0]

    if len(reactions1) != len(reactions2):
        return False

    grouped1 = _group_reactions(reactions1)  
    # e.g. {key1: [r1-0, r1-1], key2: [r1-2]}

    grouped2 = _group_reactions(reactions2)
    # e.g. {key1: [r2-1, r2-0], key2: [r2-2]}

    if grouped1.keys() != grouped2.keys():
        return False
    # Reasoning: If either of the sets has an extra key, then its
    # reactions have no counterpart in the other set.
    
    for key in grouped1.keys():
        cur_key_reactions1 = grouped1[key]  # e.g. [r1-0, r1-1]
        cur_key_reactions2 = grouped2[key]  # e.g. [r2-1, r2-0]

        if len(cur_key_reactions1) != len(cur_key_reactions2):
            return False
        
        reaction_count = len(cur_key_reactions1)

        if reaction_count == 1:
            # If there is only one reaction in the set, we can compare
            # them directly.
            if not are_equivalent_reactions(
                    cur_key_reactions1[0], cur_key_reactions2[0],
                    id_to_assembly, component_structures):
                return False
            continue

        # If there are multiple reactions in the set, we need to check
        # if there are one-to-one matches between them.

        # Cached function
        # This prevents recomputing the equivalence for the same pair.
        cached_eq = _make_cached_eq(
            cur_key_reactions1, cur_key_reactions2,
            id_to_assembly, component_structures)

        used2 = [False] * reaction_count  # e.g. [False, False]
        for i1 in range(reaction_count):  # e.g. (0 -> 1)
            for i2 in range(reaction_count):  # e.g. (0 -> 1)
                if not used2[i2] and cached_eq(i1, i2):
                    # i.e. If equivalent reactions are found...
                    used2[i2] = True  # Mark the reaction as used
                    break  # i.e. Move to the next reaction in set 1
            else:
                # If we didn't find a match for reaction i1, then the sets
                # are not equivalent.
                return False
        # If we reach here, it means all reactions in set 1 have a match
        # in set 2.
    return True


def _make_cached_eq(
        reactions1: Sequence[Reaction], reactions2: Sequence[Reaction],
        id_to_assembly: Mapping[int, Assembly],
        component_structures: Mapping[str, Component]
        ) -> Callable[[int, int], bool]:
    cache: dict[tuple[int, int], bool] = {}

    def cached_eq(index1: int, index2: int) -> bool:
        raw_key = tuple(sorted((index1, index2)))
        key = cast(tuple[int, int], raw_key)  # Only for type checking
        if key in cache:
            return cache[key]
        result = are_equivalent_reactions(
            reactions1[index1], reactions2[index2], 
            id_to_assembly, component_structures)
        cache[key] = result
        return result
    
    return cached_eq


def _group_reactions(reactions: Iterable[Reaction]) -> Mapping[tuple, list]:
    """Group reactions by their initial, entering, product, 
    and leaving assemblies.
    """
    grouped = defaultdict(list)

    grouping_key = lambda r: (
        r.init_assem_id, r.entering_assem_id,
        r.product_assem_id, r.leaving_assem_id)
    
    for reaction in reactions:
        grouped[grouping_key(reaction)].append(reaction)
    return grouped



# isort: skip_file

# Make certain subpackages available to the user as direct imports from
# the `nasap_net` namespace.
from . import assembly_connectivity
from . import assembly_equality
from . import assembly_separation
from . import aux_edge_existence
from . import hashing
from . import isomorphic_assembly_search
from . import isomorphism
from . import ligand_exchange
from . import subassembly
from . import union

# Make certain functions from some of the previous subpackages available
# to the user as direct imports from the `nasap_net` namespace.
from .assembly_connectivity import extract_connected_assemblies
from .assembly_equality import assemblies_equal
from .assembly_separation import separate_product_if_possible
from .aux_edge_existence import has_aux_edges
from .bindsite_equivalence import are_equivalent_binding_site_lists
from .bondset_sorting import sort_bondsets
from .bondset_sorting import sort_bondsets_and_bonds
from .hashing import calc_graph_hash_of_assembly
from .isomorphic_assembly_search import find_isomorphic_assembly
from .isomorphism import is_isomorphic
from .isomorphism import isomorphisms_iter
from .ligand_exchange import perform_inter_exchange
from .ligand_exchange import perform_intra_exchange
from .reaction_embedding import embed_assemblies_into_reaction
from .reaction_comparison import are_equivalent_reactions, are_equivalent_reaction_sets
from .subassembly import bond_induced_sub_assembly
from .subassembly import component_induced_sub_assembly
from .union import union_assemblies



from nasap_net.isomorphism import get_isomorphism
from nasap_net.models import Assembly, BindingSite, Bond, Component


def test_get_isomorphism():
    M = Component(kind='M', sites=[0, 1])
    L = Component(kind='L', sites=[0, 1])
    X = Component(kind='X', sites=[0])
    MLX = Assembly(
        components={'M1': M, 'L1': L, 'X1': X},
        bonds=[Bond('M1', 0, 'L1', 0), Bond('M1', 1, 'X1', 0)]
    )
    MLX_permuted = Assembly(
        components={'L2': L, 'X2': X, 'M2': M},
        bonds=[Bond('M2', 1, 'L2', 0), Bond('M2', 0, 'X2', 0)]
    )
    isom = get_isomorphism(MLX, MLX_permuted)
    assert isom.comp_id_mapping == {'M1': 'M2', 'L1': 'L2', 'X1': 'X2'}
    assert isom.binding_site_mapping == {
        BindingSite('M1', 0): BindingSite('M2', 1),
        BindingSite('M1', 1): BindingSite('M2', 0),
        BindingSite('L1', 0): BindingSite('L2', 0),
        BindingSite('L1', 1): BindingSite('L2', 1),
        BindingSite('X1', 0): BindingSite('X2', 0),
    }



from .light_to_rich import convert_light_assemblies_to_rich_ones, \
    convert_light_assembly_to_rich_one
from .rich_to_light import ConversionResult, InconsistentComponentKindError, \
    convert_assemblies_to_light_ones



from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from functools import cached_property
from types import MappingProxyType
from typing import Self

from frozendict import frozendict

from nasap_net.exceptions import IDNotSetError
from nasap_net.types import ID
from .binding_site import BindingSite
from .bond import Bond
from .component import Component


class InvalidBondError(Exception):
    def __init__(self, *, bond: Bond, msg: str = ""):
        self.bond = bond
        combined_msg = f"Invalid bond: {bond}"
        if msg:
            combined_msg += f" - {msg}"
        super().__init__(combined_msg)


@dataclass(frozen=True, init=False)
class Assembly:
    """An assembly of components connected by bonds.

    Parameters
    ----------
    components : Mapping[C, nasap_net.models.component.Component[S]]
        A mapping from component IDs to their corresponding components.
    bonds : Iterable[Bond[C, S]]
        An iterable of bonds connecting the components.

    Raises
    ------
    ValueError
        - If any bond references a non-existent component or site.
        - If a component bonds to itself.
        - If a site is used more than once.
        - If the assembly is not connected.

    Warnings
    --------
    - The assembly does not enforce connectivity; it is the user's
      responsibility to ensure that the assembly is connected as needed.
    """
    _components: frozendict[ID, Component]
    bonds: frozenset[Bond]
    _id: ID | None

    def __init__(
            self, components: Mapping[ID, Component],
            bonds: Iterable[Bond],
            *,
            id_: ID | None = None
            ):
        object.__setattr__(self, '_components', frozendict(components))
        object.__setattr__(self, 'bonds', frozenset(bonds))
        object.__setattr__(self, '_id', id_)
        self._validate()

    @property
    def id(self) -> ID:
        """Return the ID of the assembly."""
        if self._id is None:
            raise IDNotSetError("Assembly ID is not set.")
        return self._id

    @property
    def id_or_none(self) -> ID | None:
        """Return the ID of the assembly, or None if not set."""
        return self._id

    def _validate(self):
        component_keys = set(self._components.keys())
        used_sites = set()
        for bond in self.bonds:
            # Validate that the components exist
            for comp_id in bond.component_ids:
                if comp_id not in component_keys:
                    raise InvalidBondError(
                        bond=bond,
                        msg=f"Component {comp_id} not found in assembly.")

            # Validate that the sites exist in the respective components
            for site in bond.sites:
                component = self._components[site.component_id]
                if site.site_id not in component.site_ids:
                    raise InvalidBondError(
                        bond=bond,
                        msg=(
                            f"Site {site.site_id} not found in component "
                            f"{site.component_id}."))

                # Validate that the site is not already used
                if site in used_sites:
                    raise InvalidBondError(
                        bond=bond,
                        msg=f"Site {site} is already used in another bond.")
                used_sites.add(site)

    @property
    def components(self) -> Mapping[ID, Component]:
        """Return the components in the assembly as an immutable mapping."""
        return MappingProxyType(self._components)

    def _get_component_of_site(self, site: BindingSite) -> Component:
        """Return the component corresponding to the given binding site."""
        return self._components[site.component_id]

    def get_component_kind_of_site(self, site: BindingSite) -> str:
        """Return the component kind of the given binding site."""
        return self._components[site.component_id].kind

    def find_sites(
            self, *, has_bond: bool | None = None,
            component_kind: str | None = None
            ) -> frozenset[BindingSite]:
        """Return binding sites based on their bond status.
        """
        if has_bond is None and component_kind is None:
            return self._all_sites

        sites = set()
        for site in self._all_sites:
            if has_bond is not None:
                if has_bond != self.has_bond(site):
                    continue
            if component_kind is not None:
                if self.get_component_kind_of_site(site) != component_kind:
                    continue
            sites.add(site)
        return frozenset(sites)

    def has_bond(self, site: BindingSite) -> bool:
        """Check if a binding site has a bond."""
        return site in self._site_connection

    @cached_property
    def _all_sites(self) -> frozenset[BindingSite]:
        """Return all binding sites in the assembly."""
        sites = set()
        for comp_id, component in self._components.items():
            for site in component.site_ids:
                sites.add(BindingSite(component_id=comp_id, site_id=site))
        return frozenset(sites)

    @cached_property
    def _site_connection(self) -> Mapping[BindingSite, BindingSite]:
        """Return a mapping of each binding site to its connected binding site.

        Only bonded sites are included in the mapping.
        """
        connection = {}
        for bond in self.bonds:
            site1, site2 = bond.sites
            connection[site1] = site2
            connection[site2] = site1
        return MappingProxyType(connection)

    def add_bond(self, site1: BindingSite, site2: BindingSite):
        """Return a new assembly with an additional bond."""
        new_bond = Bond.from_sites(site1, site2)
        new_bonds = set(self.bonds)
        new_bonds.add(new_bond)
        return self.copy_with(bonds=new_bonds)

    def remove_bond(self, site1: BindingSite, site2: BindingSite):
        """Return a new assembly with a bond removed."""
        bond_to_remove = Bond.from_sites(site1, site2)
        new_bonds = set(self.bonds)
        new_bonds.remove(bond_to_remove)
        return self.copy_with(bonds=new_bonds)

    def copy_with(
            self,
            *,
            components: Mapping[ID, Component] | None = None,
            bonds: Iterable[Bond] | None = None,
            id_: ID | None = None,
            ) -> Self:
        """Return a copy of the assembly with optional modifications."""
        if components is None:
            components = self.components
        if bonds is None:
            bonds = self.bonds
        return self.__class__(
            components=components, bonds=bonds, id_=id_)



import pytest

from nasap_net import Assembly, Component
from nasap_net.reaction_pairing import Reaction, pair_reverse_reactions


@pytest.fixture
def components():
    return {
        'L': Component(['a', 'b']),
        'M': Component(['a', 'b']),
        'X': Component(['a']),
    }


@pytest.fixture
def assemblies():
    return {
        # MX2: X0(a)-(a)M0(b)-(a)X1
        'MX2': Assembly(
            {'M0': 'M', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'X1.a')]
        ),
        'L': Assembly({'L0': 'L'}),  # L: (a)L0(b)
        # MLX: (a)L0(b)-(a)M0(b)-(a)X0
        'MLX': Assembly(
            {'M0': 'M', 'L0': 'L', 'X0': 'X'},
            [('L0.b', 'M0.a'), ('M0.b', 'X0.a')]
        ),
        'X': Assembly({'X0': 'X'}),  # X: (a)X0
        # ML2: (a)L0(b)-(a)M0(b)-(a)L1(b)
        'ML2': Assembly(
            {'M0': 'M', 'L0': 'L', 'L1': 'L'},
            [('L0.b', 'M0.a'), ('M0.b', 'L1.a')]
        ),
        # M2LX2: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)X1
        'M2LX2': Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'X0': 'X', 'X1': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'X1.a')]
        ),
        # M2L2X: X0(a)-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)
        'M2L2X': Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L', 'X0': 'X'},
            [('X0.a', 'M0.a'), ('M0.b', 'L0.a'), ('L0.b', 'M1.a'),
             ('M1.b', 'L1.a')]
        ),
        # M2L2-ring: //-(a)M0(b)-(a)L0(b)-(a)M1(b)-(a)L1(b)-//
        'M2L2-ring': Assembly(
            {'M0': 'M', 'M1': 'M', 'L0': 'L', 'L1': 'L'},
            [('M0.b', 'L0.a'), ('L0.b', 'M1.a'), ('M1.b', 'L1.a'),
             ('L1.b', 'M0.a')]
        ),
    }


def test_M2L2_ring(components, assemblies):
    reactions = {
        '1f': Reaction(
            'MX2', 'L', 'MLX', 'X',
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a'),
        '1b': Reaction(
            'MLX', 'X', 'MX2', 'L',
            metal_bs='M0.a', leaving_bs='L0.b', entering_bs='X0.a'),
        '2f': Reaction(
            'MLX', 'L', 'ML2', 'X',
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a'),
        '2b': Reaction(
            'ML2', 'X', 'MLX', 'L',
            metal_bs='M0.a', leaving_bs='L0.b', entering_bs='X0.a'),
        '3f': Reaction(
            'MX2', 'MLX', 'M2LX2', 'X',
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a'),
        '3b': Reaction(
            'M2LX2', 'X', 'MX2', 'MLX',
            metal_bs='M0.b', leaving_bs='L0.a', entering_bs='X0.a'),
        '4f': Reaction(
            'MLX', 'MLX', 'M2L2X', 'X',
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a'),
        '4b': Reaction(
            'M2L2X', 'X', 'MLX', 'MLX',
            metal_bs='M1.a', leaving_bs='L0.b', entering_bs='X0.a'),
        '5f': Reaction(
            'MX2', 'ML2', 'M2L2X', 'X',
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a'),
        '5b': Reaction(
            'M2L2X', 'X', 'MX2', 'ML2',
            metal_bs='M0.b', leaving_bs='L0.a', entering_bs='X0.a'),
        '6f': Reaction(
            'M2LX2', 'L', 'M2L2X', 'X',
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a'),
        '6b': Reaction(
            'M2L2X', 'X', 'M2LX2', 'L',
            metal_bs='M1.b', leaving_bs='L1.a', entering_bs='X0.a'),
        '7f': Reaction(
            'M2L2X', None, 'M2L2-ring', 'X',
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L1.b'),
        '7b': Reaction(
            'M2L2-ring', 'X', 'M2L2X', None,
            metal_bs='M0.b', leaving_bs='L0.a', entering_bs='X0.a'),
        }
    result = pair_reverse_reactions(reactions, assemblies, components)
    expected = {
        '1f': '1b',
        '1b': '1f',
        '2f': '2b',
        '2b': '2f',
        '3f': '3b',
        '3b': '3f',
        '4f': '4b',
        '4b': '4f',
        '5f': '5b',
        '5b': '5f',
        '6f': '6b',
        '6b': '6f',
        '7f': '7b',
        '7b': '7f',
        }
    assert result == expected


def test_missing_pair(components, assemblies):
    """Test when one of the reverse reactions is missing."""
    reactions = {
        '1f': Reaction(
            'MX2', 'L', 'MLX', 'X',
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L0.a'),
        '1b': Reaction(
            'MLX', 'X', 'MX2', 'L',
            metal_bs='M0.a', leaving_bs='L0.b', entering_bs='X0.a'),
        '2f': Reaction(
            'MLX', 'L', 'ML2', 'X',
            metal_bs='M0.b', leaving_bs='X0.a', entering_bs='L0.a'),
        '2b': Reaction(
            'ML2', 'X', 'MLX', 'L',
            metal_bs='M0.a', leaving_bs='L0.b', entering_bs='X0.a'),
        # 3b is missing
        '3f': Reaction(
            'M2L2X', None, 'M2L2-ring', 'X',
            metal_bs='M0.a', leaving_bs='X0.a', entering_bs='L1.b'),
        }
    result = pair_reverse_reactions(reactions, assemblies, components)
    expected = {
        '1f': '1b',
        '1b': '1f',
        '2f': '2b',
        '2b': '2f',
        '3f': None,
        }
    assert result == expected



import pytest

from nasap_net.utils.cyclic_perm import cyclic_perm_to_map, cyclic_perms_to_map


def test_cyclic_perm_to_map():
    pass


def test_cyclic_perms_to_map():
    pass


if __name__ == '__main__':
    pytest.main(['-vv', __file__])



import pytest

from nasap_net.pipelines.bondset_to_assembly import validate_bond_id_to_bindsites


def test_validate_bond_id_to_bindsites_valid():
    bond_id_to_bindsites = {
        1: ['L1.b', 'M1.a'],
        2: ['M1.b', 'L2.a'],
        3: ['L2.b', 'M2.a'],
        4: ['M2.b', 'L3.a']
    }
    comp_ids = ['M1', 'M2', 'L1', 'L2', 'L3']
    # Should not raise any exception
    validate_bond_id_to_bindsites(bond_id_to_bindsites, comp_ids)


def test_validate_bond_id_to_bindsites_invalid_string():
    bond_id_to_bindsites = {
        1: ['L1.b', 123],  # Invalid: second value is not a string
    }
    comp_ids = ['M1', 'M2', 'L1', 'L2', 'L3']
    with pytest.raises(
            ValueError, 
            match='Values in "bonds_and_their_binding_sites" must be '
            'strings.'):
        validate_bond_id_to_bindsites(bond_id_to_bindsites, comp_ids)


def test_validate_bond_id_to_bindsites_same_values():
    bond_id_to_bindsites = {
        1: ['L1.b', 'L1.b'],  # Invalid: both values are the same
    }
    comp_ids = ['M1', 'M2', 'L1', 'L2', 'L3']
    with pytest.raises(
            ValueError, 
            match='Values in "bonds_and_their_binding_sites" must be '
            'different.'):
        validate_bond_id_to_bindsites(bond_id_to_bindsites, comp_ids)


def test_validate_bond_id_to_bindsites_invalid_format():
    bond_id_to_bindsites = {
        1: ['L1b', 'M1.a'],  # Invalid: first value does not have a dot
    }
    comp_ids = ['M1', 'M2', 'L1', 'L2', 'L3']
    with pytest.raises(
            ValueError,
            match='must be in the form of "comp_id.local_bindsite_id"'):
        validate_bond_id_to_bindsites(bond_id_to_bindsites, comp_ids)


def test_validate_bond_id_to_bindsites_unknown_component():
    bond_id_to_bindsites = {
        1: ['L1.b', 'X1.a'],  # Invalid: 'X1' is not in comp_ids
    }
    comp_ids = ['M1', 'M2', 'L1', 'L2', 'L3']
    with pytest.raises(
            ValueError, 
            match='Unknown component ID "X1" in '
            '"bonds_and_their_binding_sites".'):
        validate_bond_id_to_bindsites(bond_id_to_bindsites, comp_ids)


def test_validate_bond_id_to_bindsites_same_component():
    bond_id_to_bindsites = {
        1: ['M1.a', 'M1.b'],  # Invalid: both components are the same
    }
    comp_ids = ['M1', 'M2', 'L1', 'L2', 'L3']
    with pytest.raises(
            ValueError, 
            match='The two components in a bond must be different.'):
        validate_bond_id_to_bindsites(bond_id_to_bindsites, comp_ids)


if __name__ == "__main__":
    pytest.main(['-v', __file__])



from collections.abc import Callable
from typing import TypeAlias

from nasap_net import InterReactionRich, IntraReactionRich

ReactionDetailed: TypeAlias = IntraReactionRich | InterReactionRich


class ReactionClassifier:
    def __init__(
            self, 
            classification_rule: Callable[[ReactionDetailed], str]
            ) -> None:
        self.classification_rule = classification_rule

    def classify(self, reaction: ReactionDetailed) -> str:
        return self.classification_rule(reaction)



import pandas as pd
import pytest

from nasap_net import InterReaction, IntraReaction
from nasap_net.pipelines.lib import reactions_to_df
from nasap_net.pipelines.lib.reactions_df_conversion import reaction_to_dict


def test_reaction_to_dict_inter():
    inter = InterReaction(
        0, 1, 2, 3,
        'metal_bs', 'leave_bs', 'enter_bs',
        2)
    d = reaction_to_dict(inter)
    assert d == {
        'init_assem_id': 0,
        'entering_assem_id': 1,
        'product_assem_id': 2,
        'leaving_assem_id': 3,
        'metal_bs': 'metal_bs',
        'leaving_bs': 'leave_bs',
        'entering_bs': 'enter_bs',
        'duplicate_count': 2
    }


def test_reaction_to_dict_intra():
    intra = IntraReaction(
        0, 1, 2,
        'metal_bs', 'leave_bs', 'enter_bs',
        1)
    d = reaction_to_dict(intra)
    assert d == {
        'init_assem_id': 0,
        'entering_assem_id': None,
        'product_assem_id': 1,
        'leaving_assem_id': 2,
        'metal_bs': 'metal_bs',
        'leaving_bs': 'leave_bs',
        'entering_bs': 'enter_bs',
        'duplicate_count': 1
    }


def test_basic():
    reactions: list[IntraReaction | InterReaction] = [
        IntraReaction(
            0, 1, 2, 
            'metal_bs1', 'leave_bs1', 'enter_bs1', 
            1),
        InterReaction(
            3, 4, 5, None,
            'metal_bs2', 'leave_bs2', 'enter_bs2',
            2)
    ]

    df = reactions_to_df(reactions)

    assert df.columns.tolist() == [
        'init_assem_id', 'entering_assem_id', 'product_assem_id',
        'leaving_assem_id', 'metal_bs', 'leaving_bs', 'entering_bs',
        'duplicate_count'
    ]
    assert df['init_assem_id'].tolist() == [0, 3]
    assert df['entering_assem_id'].tolist() == [pd.NA, 4]
    assert df['product_assem_id'].tolist() == [1, 5]
    assert df['leaving_assem_id'].tolist() == [2, pd.NA]
    assert df['metal_bs'].tolist() == ['metal_bs1', 'metal_bs2']
    assert df['leaving_bs'].tolist() == ['leave_bs1', 'leave_bs2']
    assert df['entering_bs'].tolist() == ['enter_bs1', 'enter_bs2']
    assert df['duplicate_count'].tolist() == [1, 2]


if __name__ == '__main__':
    pytest.main(['-v', __file__])



from collections.abc import Iterable, Mapping
from itertools import combinations
from typing import TypeVar, overload

import networkx as nx
from networkx.utils import UnionFind, groups

__all__ = ['group_equivalent_nodes_or_nodesets']


@overload
def group_equivalent_nodes_or_nodesets(
        nodes_or_nodesets: Iterable[str],
        isomorphisms: Iterable[Mapping[str, str]]
        ) -> set[frozenset[str]]: ...
@overload
def group_equivalent_nodes_or_nodesets(
        nodes_or_nodesets: Iterable[tuple[str, str]],
        isomorphisms: Iterable[Mapping[str, str]]
        ) -> set[frozenset[tuple[str, str]]]: ...
@overload
def group_equivalent_nodes_or_nodesets(
        nodes_or_nodesets: Iterable[tuple[str, str, str]],
        isomorphisms: Iterable[Mapping[str, str]]
        ) -> set[frozenset[tuple[str, str, str]]]: ...
@overload
def group_equivalent_nodes_or_nodesets(
        nodes_or_nodesets: Iterable[tuple[str, ...]],
        isomorphisms: Iterable[Mapping[str, str]]
        ) -> set[frozenset[tuple[str, ...]]]: ...
def group_equivalent_nodes_or_nodesets(
        nodes_or_nodesets, isomorphisms):
    """Compute node equivalence and return it as a dictionary."""
    uf = UnionFind(nodes_or_nodesets)

    for isomorphism in isomorphisms:
        while _update_equivalence(uf, isomorphism):
            pass

    return set([frozenset(group) for group in uf.to_sets()])


def _update_equivalence(
        uf: UnionFind, isomorphism: Mapping[str, str]
        ) -> bool:
    """Update the equivalence classes of node pairs in an assembly
    based on a new isomorphism."""
    roots = set(uf[element] for element in uf)
    for root1, root2 in combinations(roots, 2):
        rev_dict = groups(uf.parents)
        group1 = rev_dict[root1]
        group2 = rev_dict[root2]
        if any(
                apply_isomorphism(element1, isomorphism)
                in group2 for element1 in group1):
            uf.union(root1, root2)
            return True
        if any(
                apply_isomorphism(element2, isomorphism)
                in group1 for element2 in group2):
            uf.union(root1, root2)
            return True
    return False


@overload
def apply_isomorphism(
        node_or_nodeset: str, isomorphism: Mapping[str, str]
        ) -> str: ...
@overload
def apply_isomorphism(
        node_or_nodeset: tuple[str, str], isomorphism: Mapping[str, str]
        ) -> tuple[str, str]: ...
@overload
def apply_isomorphism(
        node_or_nodeset: tuple[str, str, str], isomorphism: Mapping[str, str]
        ) -> tuple[str, str, str]: ...
@overload
def apply_isomorphism(
        node_or_nodeset: tuple[str, ...], isomorphism: Mapping[str, str]
        ) -> tuple[str, ...]: ...
def apply_isomorphism(
        node_or_nodeset, isomorphism):
    if isinstance(node_or_nodeset, str):
        return isomorphism[node_or_nodeset]
    return tuple(isomorphism[node] for node in node_or_nodeset)



from collections.abc import Iterable, Mapping

import yaml

from nasap_net.models import Bond, Component, LightAssembly
from nasap_net.types import ID


def load_light_assemblies(yaml_str: str) -> dict[ID, LightAssembly]:
    """Load light assemblies from a YAML string."""
    return yaml.load(yaml_str, Loader=_LightAssemblyLoader)  # type: ignore


class _LightAssemblyLoader(yaml.SafeLoader):
    component_context: Mapping[str, Component] = {}

    def ignore_aliases(self, _):
        return True


def _light_assembly_constructor(
        loader: _LightAssemblyLoader,
        node: yaml.Node,
) -> LightAssembly:
    assert isinstance(node, yaml.MappingNode)
    mapping = loader.construct_mapping(node, deep=True)
    components: dict[ID, str] = mapping['components']
    bonds: list[Bond] = [_construct_bond(b) for b in mapping['bonds']]
    assembly_id: str | None = mapping.get('id')
    return LightAssembly(
        components=components,
        bonds=bonds,
        id_=assembly_id,
    )


def _construct_bond(bonds: Iterable[ID]) -> Bond:
    comp_id1, site_id1, comp_id2, site_id2 = bonds
    return Bond(comp_id1, site_id1, comp_id2, site_id2)


yaml.add_constructor(
    '!LightAssembly', _light_assembly_constructor,
    Loader=_LightAssemblyLoader
)



from collections.abc import Iterable

from ..aux_edge import AuxEdge

__all__ = ['check_bindsites_of_aux_edges_exists']


def check_bindsites_of_aux_edges_exists(
        aux_edges: Iterable[AuxEdge], 
        bindsites: Iterable[str]) -> None:
    """Check if the binding sites of the auxiliary edges exist in the
    binding sites of the component.
    """
    for aux_edge in aux_edges:
        for bindsite in aux_edge.binding_sites:
            if bindsite not in bindsites:
                raise ValueError(
                    f'The binding site "{bindsite}" is not in '
                    f'the binding sites: {bindsites}')



from collections.abc import Mapping
from typing import TypeVar

from nasap_net.types import ID
from ..assembly import Assembly
from ..component import Component
from ..light_assembly import LightAssembly

_T = TypeVar('_T', bound=ID)

def convert_light_assembly_to_rich_one(
        light_assembly: LightAssembly,
        components: Mapping[str, Component],
) -> Assembly:
    return Assembly(
        components={
            comp_id: components[comp_kind]
            for comp_id, comp_kind in light_assembly.components.items()
        },
        bonds=light_assembly.bonds,
        id_=light_assembly.id_or_none,
    )


def convert_light_assemblies_to_rich_ones(
        light_assemblies: Mapping[_T, LightAssembly],
        components: Mapping[str, Component],
) -> dict[ID, Assembly]:
    return {
        assembly_id: convert_light_assembly_to_rich_one(light_assembly, components)
        for assembly_id, light_assembly in light_assemblies.items()
    }



from typing import Mapping

from nasap_net import Assembly, Component
from nasap_net.algorithms import isomorphisms_iter
from nasap_net.types import A, C
from ._reaction_result_generation import _generate_reaction_result
from ..models import Reaction, _MLE


class _IncorrectReactionResultError(ValueError):
    """
    Exception raised when the reproduced reaction result is inconsistent with
    the given result.
    """


def _generate_sample_rev_mle(
        reaction: Reaction[A],
        assemblies: Mapping[A, Assembly],
        components: Mapping[C, Component]
        ) -> _MLE:
    """Determine the binding site trio on the right-hand side of a reaction."""
    init_assembly = assemblies[reaction.init_assem_id]
    if reaction.entering_assem_id is None:
        entering_assembly = None
    else:
        entering_assembly = assemblies[reaction.entering_assem_id]

    reaction_result = _generate_reaction_result(
        init_assembly,
        entering_assembly,
        reaction.metal_bs,
        reaction.leaving_bs,
        reaction.entering_bs
    )

    try:
        product_isom = next(isomorphisms_iter(
            reaction_result.product_assembly,
            assemblies[reaction.product_assem_id],
            components))
    except StopIteration:
        raise _IncorrectReactionResultError(
            "The product assembly cannot be mapped to the expected one."
            ) from None

    rev_metal_bs = product_isom[reaction_result.metal_bs]
    rev_leaving_bs = product_isom[reaction_result.entering_bs]

    if reaction_result.leaving_assembly is None:
        rev_entering_bs = product_isom[reaction_result.leaving_bs]
    else:
        assert reaction.leaving_assem_id is not None
        try:
            leaving_isom = next(isomorphisms_iter(
                reaction_result.leaving_assembly,
                assemblies[reaction.leaving_assem_id],
                components))
        except StopIteration:
            raise _IncorrectReactionResultError(
                "The leaving assembly cannot be mapped to the expected one."
                ) from None
        rev_entering_bs = leaving_isom[reaction_result.leaving_bs]
    return _MLE(rev_metal_bs, rev_leaving_bs, rev_entering_bs)



from collections.abc import Iterator

from networkx.algorithms.isomorphism import (GraphMatcher,
                                             categorical_node_match)

from nasap_net import Assembly

__all__ = ['rough_isomorphisms_iter']


def rough_isomorphisms_iter(
        assem1: Assembly, assem2: Assembly
        ) -> Iterator[dict[str, str]]:
    """Find all isomorphisms between two "rough_g_snapshot" graphs of 
    assemblies.

    Parameters
    ----------
    assem1 : Assembly
        The first assembly.
    assem2 : Assembly
        The second assembly.

    Yields
    ------
    dict[str, str]
        A mapping from component ID (e.g., 'M1', 'L1') in the first assembly
        to component ID in the second assembly.
    """
    node_match = categorical_node_match('component_kind', None)
    GM = GraphMatcher(
        assem1.rough_g_snapshot, assem2.rough_g_snapshot, node_match=node_match)
    yield from GM.isomorphisms_iter()
